#!/bin/bash
# $1 is the kernel build dir
# $2 is the kernel source dir (might be the same)
# $3 is the generated check file

if [ $# -lt 3 ] ; then
	echo "$0 needs KSRC, KDIR and OUTPUT as arguments"
	exit -1
fi

KSRC=$1
KDIR=$2
OUTPUT=$3

# Find where the headers are (to avoid grepping at both places).
# Do not check for autoconf.h or version.h since these are in
# both the source and the build directory.
HEADERS=
if [ -f ${KSRC}/include/linux/kernel.h ] ; then
	HEADERS=$KSRC
else if [ -f ${KDIR}/include/linux/kernel.h ] ; then
	HEADERS=$KDIR
fi fi

# check that we found kernel headers
if [ -z ${HEADERS} ] ; then
	echo "Cannot find include/linux/kernel.h in ${KSRC} or ${KDIR}"
	exit -1
fi
echo "Using kernel headers in ${HEADERS}"

# generate the output file
rm -f ${OUTPUT}

# add the header
echo "#ifndef __MYRI10GE_CHECKS_H__" >> ${OUTPUT}
echo "#define __MYRI10GE_CHECKS_H__ 1" >> ${OUTPUT}
echo "" >> ${OUTPUT}

# what command line was used to generate with file
echo "/*" >> ${OUTPUT}
echo " * This file has been generated with check_headers.sh on "`date` >> ${OUTPUT}
echo " * It has been called with:" >> ${OUTPUT}
echo " *   KSRC=${KSRC}" >> ${OUTPUT}
echo " *   KDIR=${KDIR}" >> ${OUTPUT}
echo " * It checked kernel headers in ${HEADERS}/include/" >> ${OUTPUT}
echo " */" >> ${OUTPUT}
echo "" >> ${OUTPUT}

# pci_save/restore_state lost its second argument in 2.6.10
grep "pci_save_state *(.*dev, .*buffer)" ${HEADERS}/include/linux/pci.h  > /dev/null \
  && echo "#define MYRI10GE_HAVE_PRIVATE_PM_STATE 1" >> ${OUTPUT} || true

# pm_message_t appeared in 2.6.11
grep pm_message_t ${HEADERS}/include/linux/pm.h > /dev/null \
  && echo "#define MYRI10GE_HAVE_PM_MESSAGE_T 1" >> ${OUTPUT} || true

# skb_linearize had a gfp argument before 2.6.18
grep "skb_linearize *(.*, .* gfp)" ${HEADERS}/include/linux/skbuff.h > /dev/null \
  && echo "#define MYRI10GE_SKB_LINEARIZE_HAS_GFP 1" >> ${OUTPUT} || true

# skb_padto returned a sk_buff before 2.6.18
grep "sk_buff \*skb_padto *(" ${HEADERS}/include/linux/skbuff.h > /dev/null \
  && echo "#define MYRI10GE_SKB_PADTO_RETURNS_SKB 1" >> ${OUTPUT} || true

# netdev_alloc_skb deprecates dev_alloc_skb in 2.6.18
grep "netdev_alloc_skb *(" ${HEADERS}/include/linux/skbuff.h > /dev/null \
  && echo "#define MYRI10GE_HAVE_NETDEV_ALLOC_SKB 1" >> ${OUTPUT} || true

# the irq handler losts its regs argument in 2.6.19
grep "irqreturn_t.*(.*)(int, void \*, struct pt_regs \*)" ${HEADERS}/include/linux/interrupt.h > /dev/null \
  && echo "#define MYRI10GE_HAVE_IRQ_HANDLER_REGS 1" >> ${OUTPUT} || true

# add the footer
echo "" >> ${OUTPUT}
echo "#endif /* __MYRI10GE_CHECKS_H__ */" >> ${OUTPUT}
