/*************************************************************************
 * myri10ge.c: Myricom Myri-10G Ethernet driver.
 *
 * Copyright (C) 2005 - 2007 Myricom, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Myricom, Inc. nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * If the eeprom on your board is not recent enough, you will need to get a
 * newer firmware image at:
 *   http://www.myri.com/scs/download-Myri10GE.html
 *
 * Contact Information:
 *   <help@myri.com>
 *   Myricom, Inc., 325N Santa Anita Avenue, Arcadia, CA 91006
 *************************************************************************/

#ifndef LINUX_KERNEL_SPECIFIC
static const char __idstring[] = "$Id: myri10ge.c,v 1.260 2007/07/16 14:12:48 gallatin Exp $";
#ifndef AUTOCONF_INCLUDED
#include <linux/config.h>
#endif
#endif /* LINUX_KERNEL_SPECIFIC */
#include <linux/tcp.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/string.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/dma-mapping.h>
#include <linux/etherdevice.h>
#include <linux/if_ether.h>
#include <linux/if_vlan.h>
#include <linux/ip.h>
#include <linux/inet.h>
#include <linux/in.h>
#include <linux/ethtool.h>
#include <linux/firmware.h>
#include <linux/delay.h>
#include <linux/version.h>
#include <linux/timer.h>
#include <linux/vmalloc.h>
#include <linux/crc32.h>
#ifdef LINUX_KERNEL_SPECIFIC
#include <linux/moduleparam.h>
#include <linux/io.h>
#include <linux/log2.h>
#endif /* LINUX_KERNEL_SPECIFIC */
#include <net/checksum.h>
#if MYRI10GE_LRO
#include <net/ip.h>
#include <net/tcp.h>
#endif
#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/processor.h>
#ifdef CONFIG_MTRR
#include <asm/mtrr.h>
#endif

#ifndef LINUX_KERNEL_SPECIFIC
#include "myri10ge_hal.h"
#include "myri10ge_checks.h"

#include "../firmware/myri10ge_mcp.h"
#include "../firmware/mcp_gen_header.h"
#if MYRI10GE_BUILTIN_FW
#include "../firmware/eth_z8e.h"
#include "../firmware/ethp_z8e.h"
#include <linux/zlib.h>
#endif /* MYRI10GE_BUILTIN_FW */
#else /* LINUX_KERNEL_SPECIFIC */
#include "myri10ge_mcp.h"
#include "myri10ge_mcp_gen_header.h"

#define MYRI10GE_VERSION_STR MYRI10GE_VERSION_STR
#endif /* LINUX_KERNEL_SPECIFIC */

MODULE_DESCRIPTION("Myricom 10G driver (10GbE)");
MODULE_AUTHOR("Maintainer: help@myri.com");
MODULE_VERSION(MYRI10GE_VERSION_STR);
MODULE_LICENSE("Dual BSD/GPL");

#define MYRI10GE_MAX_ETHER_MTU 9014

#define MYRI10GE_ETH_STOPPED 0
#define MYRI10GE_ETH_STOPPING 1
#define MYRI10GE_ETH_STARTING 2
#define MYRI10GE_ETH_RUNNING 3
#define MYRI10GE_ETH_OPEN_FAILED 4

#define MYRI10GE_EEPROM_STRINGS_SIZE 256
#define MYRI10GE_MAX_SEND_DESC_TSO ((65536 / 2048) * 2)

#define MYRI10GE_NO_CONFIRM_DATA htonl(0xffffffff)
#define MYRI10GE_NO_RESPONSE_RESULT 0xffffffff

#ifdef LINUX_KERNEL_SPECIFIC
#define MYRI10GE_ALLOC_ORDER 0
#else
#define MYRI10GE_NAPI
#endif
#define MYRI10GE_ALLOC_SIZE ((1 << MYRI10GE_ALLOC_ORDER) * PAGE_SIZE)
#define MYRI10GE_MAX_FRAGS_PER_FRAME (MYRI10GE_MAX_ETHER_MTU/MYRI10GE_ALLOC_SIZE + 1)

#ifndef LINUX_KERNEL_SPECIFIC
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16) && MYRI10GE_ALLOC_ORDER != 0
#error "High order allocations are supported only on Linux 2.6.16 and newer"
#endif 
#endif /* LINUX_KERNEL_SPECIFIC */

#if MYRI10GE_RX_SKBS
#if MYRI10GE_ALLOC_ORDER != 0
#warning "High order allocations are only meaninfgul if MYRI10GE_RX_SKBS is 0"
#endif /* MYRI10GE_ALLOC_ORDER */
#if MYRI10GE_LRO != 0
#warning "LRO can only be used if MYRI10GE_RX_SKBS is 0"
#undef MYRI10GE_LRO
#define MYRI10GE_LRO 0
#endif /* MYRI10GE_LRO */
#endif /* MYRI10GE_RX_SKBS */

struct myri10ge_rx_buffer_state {
#if MYRI10GE_RX_SKBS
	struct sk_buff *skb;
#else
	struct page *page;
	int page_offset;
#endif
	DECLARE_PCI_UNMAP_ADDR(bus)
	DECLARE_PCI_UNMAP_LEN(len)
};

struct myri10ge_tx_buffer_state {
	struct sk_buff *skb;
	int last;
	DECLARE_PCI_UNMAP_ADDR(bus)
	DECLARE_PCI_UNMAP_LEN(len)
};

struct myri10ge_cmd {
	u32 data0;
	u32 data1;
	u32 data2;
};

struct myri10ge_rx_buf {
	struct mcp_kreq_ether_recv __iomem *lanai;	/* lanai ptr for recv ring */
	u8 __iomem *wc_fifo;			/* w/c rx dma addr fifo address */
	struct mcp_kreq_ether_recv *shadow;	/* host shadow of recv ring */
	struct myri10ge_rx_buffer_state *info;
	struct page *page;
	dma_addr_t bus;	
	int page_offset;
	int cnt;
	int fill_cnt;
	int alloc_fail;
	int mask;			/* number of rx slots -1 */
	int watchdog_needed;
};

struct myri10ge_tx_buf {
	struct mcp_kreq_ether_send __iomem *lanai;	/* lanai ptr for sendq */
	u8 __iomem *wc_fifo;		/* w/c send fifo address */
	struct mcp_kreq_ether_send *req_list;	/* host shadow of sendq */
	char *req_bytes;
	struct myri10ge_tx_buffer_state *info;
	int mask;			/* number of transmit slots -1	*/
	int boundary;			/* boundary transmits cannot cross*/
	int req ____cacheline_aligned;	/* transmit slots submitted	*/
	int pkt_start;			/* packets started */
	int done ____cacheline_aligned;	/* transmit slots completed	*/
	int pkt_done;			/* packets completed */
};

#if MYRI10GE_LRO
struct  myri10ge_lro_packet {
	struct hlist_node lro_node;
	struct sk_buff *skb;
	int	timestamp;
	__u32	tsval;
	__u32	tsecr;
	__u32	source_ip;
	__u32	dest_ip;
	__u32	next_seq;
	__u32	ack_seq;
	__wsum	data_csum;
	__u16	window;
	__u16	source_port;
	__u16	dest_port;
	__u16	append_cnt;
	__u16	mss;
	__u16	vlan_tci;
};

#endif /* MYRI10GE_LRO */

struct myri10ge_rx_done {
	struct mcp_slot *entry;
	dma_addr_t bus;
	int cnt;
	int idx;
#if MYRI10GE_LRO
	struct hlist_head lro_active;
	struct hlist_head lro_free;
#endif /* MYRI10GE_LRO */
};


struct myri10ge_priv {
	int running;			/* running? 		*/
	int csum_flag;			/* rx_csums? 		*/
	struct myri10ge_tx_buf tx;	/* transmit ring 	*/
	struct myri10ge_rx_buf rx_small;
	struct myri10ge_rx_buf rx_big;
	struct myri10ge_rx_done rx_done;
	int small_bytes;
	int big_bytes;
#if MYRI10GE_LRO
	int lro_flushed;
	int lro_queued;
	int lro_too_many_streams;
	int lro_bad_csum;
#endif /* MYRI10GE_LRO */
	struct net_device *dev;
	struct net_device_stats stats;
	u8 __iomem *sram;
	int sram_size;
	unsigned long board_span;
	unsigned long iomem_base;
	__be32 __iomem *irq_claim;
	__be32 __iomem *irq_deassert;
	char *mac_addr_string;
	struct mcp_cmd_response *cmd;
	dma_addr_t cmd_bus;
	struct mcp_irq_data *fw_stats;
	dma_addr_t fw_stats_bus;
	struct pci_dev *pdev;
	int msi_enabled;
	__be32 link_state;
	unsigned int rdma_tags_available;
	int intr_coal_delay;
	__be32 __iomem *intr_coal_delay_ptr;
	int mtrr;
	int wc_enabled;
	int wake_queue;
	int stop_queue;
	int down_cnt;
	wait_queue_head_t down_wq;
	struct work_struct watchdog_work;
	struct timer_list watchdog_timer;
	int watchdog_tx_done;
	int watchdog_tx_req;
	int watchdog_resets;
	int tx_linearized;
	int pause;
	char *fw_name;
	char eeprom_strings[MYRI10GE_EEPROM_STRINGS_SIZE];
	char fw_version[128];
	int fw_ver_major;
	int fw_ver_minor;
	int fw_ver_tiny;
	int adopted_rx_filter_bug;
	u8	mac_addr[6];	/* eeprom mac address */
	unsigned long serial_number;
	int vendor_specific_offset;
	int fw_multicast_support;
#ifndef LINUX_KERNEL_SPECIFIC
	u32 devctl;
	u32 msi_addr_low;
	u32 msi_addr_high;
	u16 msi_data_32;
	u16 msi_data_64;
	u16 msi_flags;
#endif
#ifdef MYRI10GE_HAVE_PRIVATE_PM_STATE
	u32 pm_state[16];
#endif
	u32 read_dma;
	u32 write_dma;
	u32 read_write_dma;
	u32 link_changes;
	u32 msg_enable;
};

static char *myri10ge_fw_unaligned = "myri10ge_ethp_z8e.dat";
static char *myri10ge_fw_aligned = "myri10ge_eth_z8e.dat";

static char *myri10ge_fw_name = NULL;
module_param(myri10ge_fw_name, charp, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(myri10ge_fw_name, "Firmware image name\n");

static int myri10ge_ecrc_enable = 1;
module_param(myri10ge_ecrc_enable, int, S_IRUGO);
MODULE_PARM_DESC(myri10ge_ecrc_enable, "Enable Extended CRC on PCI-E\n");

static int myri10ge_max_intr_slots = 1024;
module_param(myri10ge_max_intr_slots, int, S_IRUGO);
MODULE_PARM_DESC(myri10ge_max_intr_slots, "Interrupt queue slots\n");

static int myri10ge_small_bytes = -1;	/* -1 == auto */
module_param(myri10ge_small_bytes, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(myri10ge_small_bytes, "Threshold of small packets\n");

#ifdef MYRI10GE_HAVE_MSI
#ifdef LINUX_KERNEL_SPECIFIC
static int myri10ge_msi = 1;	/* enable msi by default */
#else
static int myri10ge_msi = -1;	/* 0: off, 1:on, otherwise auto */
#endif
module_param(myri10ge_msi, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(myri10ge_msi, "Enable Message Signalled Interrupts\n");
#endif /* MYRI10GE_HAVE_MSI */

#ifdef LINUX_KERNEL_SPECIFIC
static int myri10ge_intr_coal_delay = 75;
#else
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15))
static int myri10ge_intr_coal_delay = 75;
#else
static int myri10ge_intr_coal_delay = 25;
#endif
#endif /* LINUX_KERNEL_SPECIFIC */
module_param(myri10ge_intr_coal_delay, int, S_IRUGO);
MODULE_PARM_DESC(myri10ge_intr_coal_delay, "Interrupt coalescing delay\n");

static int myri10ge_flow_control = 1;
module_param(myri10ge_flow_control, int, S_IRUGO);
MODULE_PARM_DESC(myri10ge_flow_control, "Pause parameter\n");

static int myri10ge_deassert_wait = 1;
module_param(myri10ge_deassert_wait, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(myri10ge_deassert_wait, "Wait when deasserting legacy interrupts\n");

static int myri10ge_force_firmware = 0;
module_param(myri10ge_force_firmware, int, S_IRUGO);
MODULE_PARM_DESC(myri10ge_force_firmware, "Force firmware to assume aligned completions\n");

static int myri10ge_initial_mtu = MYRI10GE_MAX_ETHER_MTU - ETH_HLEN;
module_param(myri10ge_initial_mtu, int, S_IRUGO);
MODULE_PARM_DESC(myri10ge_initial_mtu, "Initial MTU\n");

#ifndef LINUX_KERNEL_SPECIFIC
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)
static int myri10ge_vlan_csum_fixup = 1;
#else
static int myri10ge_vlan_csum_fixup = 0;
#endif
module_param(myri10ge_vlan_csum_fixup, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(myri10ge_vlan_csum_fixup, "Force VLAN Checksum fixup\n");
#endif /* LINUX_KERNEL_SPECIFIC */

static int myri10ge_napi_weight = 64;
module_param(myri10ge_napi_weight, int, S_IRUGO);
MODULE_PARM_DESC(myri10ge_napi_weight, "Set NAPI weight\n");

static int myri10ge_watchdog_timeout = 1;
module_param(myri10ge_watchdog_timeout, int, S_IRUGO);
MODULE_PARM_DESC(myri10ge_watchdog_timeout, "Set watchdog timeout\n");

static int myri10ge_max_irq_loops = 1048576;
module_param(myri10ge_max_irq_loops, int, S_IRUGO);
MODULE_PARM_DESC(myri10ge_max_irq_loops, "Set stuck legacy IRQ detection threshold\n");

#define MYRI10GE_MSG_DEFAULT NETIF_MSG_LINK

static int myri10ge_debug = -1;	/* defaults above */
module_param(myri10ge_debug, int, 0);
MODULE_PARM_DESC(myri10ge_debug, "Debug level (0=none,...,16=all)");

#ifndef LINUX_KERNEL_SPECIFIC
#if MYRI10GE_LRO
static int myri10ge_lro = 8;
#else
static int myri10ge_lro = 0;
#endif /* LINUX_KERNEL_SPECIFIC */
module_param(myri10ge_lro, int, S_IRUGO);
MODULE_PARM_DESC(myri10ge_lro, "Enable large N receive offload queues\n");
#endif /* MYRI10GE_LRO */

static int myri10ge_fill_thresh = 256;
module_param(myri10ge_fill_thresh, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(myri10ge_fill_thresh, "Number of empty rx slots allowed\n");

static int myri10ge_reset_recover = 1;
#ifndef LINUX_KERNEL_SPECIFIC
module_param(myri10ge_reset_recover, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(myri10ge_reset_recover, "Number of recoveries allowed from NIC hw reset\n");

static int myri10ge_bus = -1;
module_param(myri10ge_bus, int, S_IRUGO);
MODULE_PARM_DESC(myri10ge_bus, "Only consider devices on this PCI bus\n");
#endif

static int myri10ge_wcfifo = 0;
module_param(myri10ge_wcfifo, int, S_IRUGO);
MODULE_PARM_DESC(myri10ge_wcfifo, "Enable WC Fifo when WC is enabled\n");

#if MYRI10GE_HAVE_PAT
static int myri10ge_pat_failed = 0;
#endif
#define MYRI10GE_FW_OFFSET 1024*1024
#define MYRI10GE_HIGHPART_TO_U32(X) \
(sizeof (X) == 8) ? ((u32)((u64)(X) >> 32)) : (0)
#define MYRI10GE_LOWPART_TO_U32(X) ((u32)(X))

#ifdef LINUX_KERNEL_SPECIFIC
#define myri10ge_pio_copy(to,from,size) __iowrite64_copy(to,from,size/8)
#endif /* LINUX_KERNEL_SPECIFIC */

static void myri10ge_set_multicast_list(struct net_device *dev);

static inline void put_be32(__be32 val, __be32 __iomem *p)
{
	__raw_writel((__force __u32)val, (__force void __iomem *)p);
}

#ifndef LINUX_KERNEL_SPECIFIC
#ifdef MYRI10GE_HAVE_MSI
static int
myri10ge_hyper_msi_cap_on(struct pci_dev *pdev)
{
	u8 cap_off;
	int nbcap = 0;

	cap_off = PCI_CAPABILITY_LIST - 1;
	/* go through all caps looking for a hypertransport msi mapping */
	while (pci_read_config_byte(pdev, cap_off + 1, &cap_off) == 0 &&
	       nbcap++ <= 256 / 4) {
		u32 cap_hdr;
		if (cap_off == 0 || cap_off == 0xff)
			break;
		cap_off &= 0xfc;
		/* cf hypertransport spec, msi mapping section */
		if (pci_read_config_dword(pdev, cap_off, &cap_hdr) == 0
		    && (cap_hdr & 0xff) == 8 /* hypertransport cap */
		    && (cap_hdr & 0xf8000000) == 0xa8000000 /* msi mapping */) {
			if (cap_hdr & 0x10000) /* msi mapping cap enabled */
				/* MSI present and enabled */
				return 1;
			else
				/* MSI present and disabled */
				return 0;
		}
	}
	/* MSI absent */
	return 0;
}

static int
myri10ge_use_msi(struct pci_dev *pdev)
{
	if (myri10ge_msi == 1 || myri10ge_msi == 0)
		return myri10ge_msi;

	/*  find root complex for our device */
	while (pdev->bus && pdev->bus->self)
		pdev = pdev->bus->self;

	/* go for it if chipset is intel, or has hypertransport msi cap */
	if (pdev->vendor == PCI_VENDOR_ID_INTEL
	    || myri10ge_hyper_msi_cap_on(pdev))
		return 1;

	/*  check if main chipset device has hypertransport msi cap */
	pdev = pci_find_slot(pdev->bus->number, 0);
	if (pdev && myri10ge_hyper_msi_cap_on(pdev))
		return 1;

	/* default off */
	return 0;
}
#endif /* MYRI10GE_HAVE_MSI */

#if MYRI10GE_HAVE_PAT
#define MYRI10GE_DEFAULT_PAT 0x7040600070406ULL
#define MYRI10GE_ENABLED_PAT6 0x1040600070406ULL

#ifdef CONFIG_X86_64
#ifndef MAXMEM
#include <asm/e820.h>
#endif /*MAXMEM*/
#endif /*CONFIG_X86_64*/
/* 
 * This function is used to work around a quirk of the linux kernel which
 * would otherwise cause our driver to leak 16MB of ram per interface
 * when PAT write-combining is used.
 *
 * Early in the boot process, Linux maps linearly all physical space
 * at: [PAGE_OFFSET, PAGE_OFFSET + <end-of-usable-physical-space]
 * ioremap() gives a new mapping for the same physical space in a
 * different virtual region (the "vmalloc" region). Linux tries to
 * keep the same attributes for the two virtual mapping of the same
 * physical space through the clumsy change_page_attr(). After
 * ioremap() has more or less established the new mapping,
 * change_page_attr() is called on the corresponding interval of the
 * "linear mapping" to fix it, and this part is confused by the
 * presence of the PAT (_PAGE_PSE) bit in the pte entries and leaks
 * memory.  This function is designed to be called prior to iounmap()
 * to clear the _PAGE_PSE bits in the linear mapping, and eliminate
 * this leak of vmalloc space.
*/

static void
myri10ge_cleanup_linear_map(struct myri10ge_priv *mgp)
{
	pgd_t *pgd;
#ifdef PUD_SHIFT
	pud_t *pud;
#else
	pgd_t *pud;
#endif
	pmd_t *pmd;
	pte_t *pte;
	unsigned long offset;
	unsigned long addr;
	struct pci_dev *pdev = mgp->pdev;
	int warning = 0;

	if (mgp->wc_enabled != 2)
		return;

	addr = pci_resource_start(pdev, 0);
	if (addr == 0)
		return;
#ifdef CONFIG_X86_64
	if (addr >= MAXMEM)
		return;
#else
	if (addr >= virt_to_phys(high_memory))
		return;
#endif
	for (offset = 0;offset < pci_resource_len(pdev, 0);
	     offset += PAGE_SIZE) {
		addr = (unsigned long)__va(pci_resource_start(pdev, 0) + offset);
		pgd = pgd_offset_k(addr);
		if (!pgd_present(*pgd)) {
			printk("pgd not present\n");
			return;
		}
#ifdef PUD_SHIFT
		pud = pud_offset(pgd, addr);
		if (!pud_present(*pud)) {
			printk("pud not present\n");
			return;
		}
#else
		pud = pgd;
#endif
		pmd = pmd_offset(pud, addr);
		if (!pmd_present(*pmd)) {
			printk("pmd not present\n");
			return;
		}
		if (pmd_large(*pmd)) {
			printk("pmd large\n");
			return;
		}
		pte = pte_offset_kernel(pmd, addr);
		if (pte_present(*pte) && (pte_val(*pte) & _PAGE_PSE)) {
#ifdef CONFIG_X86_64
			pte->pte &= ~_PAGE_PSE;
#else
			pte->pte_low &= ~_PAGE_PSE;
#endif
		} else if (!warning++) {
			if pte_present(*pte)
				printk("%p at offset 0x%lx 0x%lx!\n",
				       pte, offset,
				       (unsigned long)pte_val(*pte));
			else
				printk("pte not present at offset 0x%lx!\n",
				       offset);
		}
	}
}

static void
myri10ge_enable_pat(void *info)
{
	/* use PAT 6 */
	static spinlock_t lock = SPIN_LOCK_UNLOCKED;
	u64 val;
	static int warned = 0;

	unsigned id = smp_processor_id();
	uint32_t low, high;
	
	spin_lock(&lock);
	rdmsr(IA32_MSR_CR_PAT, low, high);
	val = ((u64)high << 32ULL) | low;
	if (val != MYRI10GE_DEFAULT_PAT && val != MYRI10GE_ENABLED_PAT6) {
		if (!warned) {
			warned = 1;
			printk(KERN_WARNING "CPU%d: existing PAT "
			       "has non-default value =  0x%x%08x\n", 
			       id,high, low);
			printk(KERN_WARNING "PAT not enabled!\n");
		}
		goto abort_with_lock;
	}
	if ((high & 0xff0000) != 0x010000) {
		high = (high & 0xff00ffff) | 0x00010000;
		wrmsr(IA32_MSR_CR_PAT, low, high);
		rdmsr(IA32_MSR_CR_PAT, low, high);
		val = ((u64)high << 32ULL) | low;
	}

abort_with_lock:
	if (val != MYRI10GE_ENABLED_PAT6)
		myri10ge_pat_failed = 1;

	spin_unlock(&lock);
}
#else
#define myri10ge_cleanup_linear_map(x)
#endif

#endif /* LINUX_KERNEL_SPECIFIC */


static int
myri10ge_send_cmd(struct myri10ge_priv *mgp, u32 cmd,
		  struct myri10ge_cmd *data, int atomic)
{
	struct mcp_cmd *buf;
	char buf_bytes[sizeof(*buf) + 8];
	struct mcp_cmd_response *response = mgp->cmd;
	char __iomem *cmd_addr = mgp->sram + MXGEFW_ETH_CMD;
	u32 dma_low, dma_high, result, value;
	int sleep_total = 0;

	/* ensure buf is aligned to 8 bytes */
	buf = (struct mcp_cmd *) ALIGN((unsigned long) buf_bytes, 8);

	buf->data0 = htonl(data->data0);
	buf->data1 = htonl(data->data1);
	buf->data2 = htonl(data->data2);
	buf->cmd = htonl(cmd);
	dma_low = MYRI10GE_LOWPART_TO_U32(mgp->cmd_bus);
	dma_high = MYRI10GE_HIGHPART_TO_U32(mgp->cmd_bus);

	buf->response_addr.low = htonl(dma_low);
	buf->response_addr.high = htonl(dma_high);
	response->result = htonl(MYRI10GE_NO_RESPONSE_RESULT);
	mb();
	myri10ge_pio_copy(cmd_addr, buf, sizeof (*buf));

	/* wait up to 15ms. Longest command is the DMA benchmark,
	 * which is capped at 5ms, but runs from a timeout handler
	 * that runs every 7.8ms. So a 15ms timeout leaves us with
	 * a 2.2ms margin
	 */
	if (atomic) {
		/* if atomic is set, do not sleep,
		 * and try to get the completion quickly
		 * (1ms will be enough for those commands) */
		for (sleep_total = 0;
		     sleep_total < 1000
		     && response->result == htonl(MYRI10GE_NO_RESPONSE_RESULT);
		     sleep_total += 10)
			udelay(10);
	} else {
		/* use msleep for most command */
		for (sleep_total = 0;
		     sleep_total < 15
		     && response->result == htonl(MYRI10GE_NO_RESPONSE_RESULT);
		     sleep_total++)
			myri10ge_msleep(1);
	}

	result = ntohl(response->result);
	value = ntohl(response->data);
	if (result != MYRI10GE_NO_RESPONSE_RESULT) {
		if (result == 0) {
			data->data0 = value;
			return 0;
		} else if (result == MXGEFW_CMD_UNKNOWN) {
			return -ENOSYS;
		} else if (result == MXGEFW_CMD_ERROR_UNALIGNED) {
			return -E2BIG;
		} else {
			dev_err(&mgp->pdev->dev,
				"command %d failed, result = %d\n",
				cmd, result);
			return -ENXIO;
		}
	}

	dev_err(&mgp->pdev->dev, "command %d timed out, result = %d\n",
	       cmd, result);
	return -EAGAIN;
}


/*
 * The eeprom strings on the lanaiX have the format
 * SN=x\0
 * MAC=x:x:x:x:x:x\0
 * PT:ddd mmm xx xx:xx:xx xx\0
 * PV:ddd mmm xx xx:xx:xx xx\0
 */
static int
myri10ge_read_mac_addr(struct myri10ge_priv *mgp)
{
	char *ptr, *limit;
	int i;

	ptr = mgp->eeprom_strings;
	limit = mgp->eeprom_strings + MYRI10GE_EEPROM_STRINGS_SIZE;

	while (*ptr != '\0' && ptr < limit) {
		if (memcmp(ptr, "MAC=", 4) == 0) {
			ptr += 4;
			mgp->mac_addr_string = ptr;
			for (i = 0; i < 6; i++) {
				if ((ptr + 2) > limit)
					goto abort;
				mgp->mac_addr[i] = simple_strtoul(ptr, &ptr, 16);
				ptr += 1;
			}
		}
		if (memcmp((const void *) ptr, "SN=", 3) == 0) {
			ptr += 3;
			mgp->serial_number = simple_strtoul(ptr, &ptr, 10);
		}
		while (ptr < limit && *ptr++);
	}

	return 0;

 abort:
	dev_err(&mgp->pdev->dev, "failed to parse eeprom_strings\n");
	return -ENXIO;
}

/*
 * Enable or disable periodic RDMAs from the host to make certain
 * chipsets resend dropped PCIe messages
 */

static void
myri10ge_dummy_rdma(struct myri10ge_priv *mgp, int enable)
{
	char __iomem *submit;
	__be32 buf[16];
	u32 dma_low, dma_high;
	int i;

	/* clear confirmation addr */
	mgp->cmd->data = 0;
	mb();

	/* send a rdma command to the PCIe engine, and wait for the
	 * response in the confirmation address.  The firmware should
	 * write a -1 there to indicate it is alive and well
	 */
	dma_low = MYRI10GE_LOWPART_TO_U32(mgp->cmd_bus);
	dma_high = MYRI10GE_HIGHPART_TO_U32(mgp->cmd_bus);

	buf[0] = htonl(dma_high); 	/* confirm addr MSW */
	buf[1] = htonl(dma_low); 	/* confirm addr LSW */
	buf[2] = MYRI10GE_NO_CONFIRM_DATA;	/* confirm data */
	buf[3] = htonl(dma_high); 	/* dummy addr MSW */
	buf[4] = htonl(dma_low); 	/* dummy addr LSW */
	buf[5] = htonl(enable);		/* enable? */

	submit = mgp->sram + MXGEFW_BOOT_DUMMY_RDMA;

	myri10ge_pio_copy(submit, &buf, sizeof (buf));
	for (i = 0; mgp->cmd->data != MYRI10GE_NO_CONFIRM_DATA && i < 20; i++)
		myri10ge_msleep(1);
	if (mgp->cmd->data != MYRI10GE_NO_CONFIRM_DATA)
		dev_err(&mgp->pdev->dev, "dummy rdma %s failed\n",
			(enable ? "enable" : "disable"));
}

static int
myri10ge_validate_firmware(struct myri10ge_priv *mgp,
			   struct mcp_gen_header *hdr)
{
	struct device *dev = &mgp->pdev->dev;


	/* check firmware type */
	if (ntohl(hdr->mcp_type) != MCP_TYPE_ETH) {
		dev_err(dev, "Bad firmware type: 0x%x\n",
			ntohl(hdr->mcp_type));
		return -EINVAL;
	}

	/* save firmware version for ethtool */
	strncpy(mgp->fw_version, hdr->version, sizeof (mgp->fw_version));

	sscanf(mgp->fw_version, "%d.%d.%d", &mgp->fw_ver_major,
	       &mgp->fw_ver_minor, &mgp->fw_ver_tiny);

	if (!(mgp->fw_ver_major == MXGEFW_VERSION_MAJOR
	      && mgp->fw_ver_minor == MXGEFW_VERSION_MINOR)) {
		dev_err(dev, "Found firmware version %s\n",
			mgp->fw_version);
		dev_err(dev, "Driver needs %d.%d\n", MXGEFW_VERSION_MAJOR,
			MXGEFW_VERSION_MINOR);
		return -EINVAL;
	}
	return 0;
}

#if MYRI10GE_BUILTIN_FW
static int
myri10ge_load_builtin_firmware(struct myri10ge_priv *mgp, u32 *size)
{
	z_stream zs;
	void *inflate_buffer;
	const unsigned char *mcp;
	struct device *dev = &mgp->pdev->dev;
	struct mcp_gen_header *hdr;
	size_t hdr_offset;
	int i, zerr, status = -ENXIO;
	unsigned int mcp_uncompressed_len, mcp_len;
	unsigned crc, reread_crc;


	if (!strcmp(mgp->fw_name, myri10ge_fw_aligned)) {
		mcp_uncompressed_len = eth_z8e_uncompressed_length;
		mcp_len = eth_z8e_length;
		mcp = eth_z8e;
	} else if (!strcmp(mgp->fw_name, myri10ge_fw_unaligned)) {
		mcp_uncompressed_len = ethp_z8e_uncompressed_length;
		mcp_len = ethp_z8e_length;
		mcp = ethp_z8e;
	} else {
		dev_err(dev, "No %s firmware built in\n", mgp->fw_name);
		return -ENXIO;
	}

	inflate_buffer = vmalloc(mcp_uncompressed_len);
	if (inflate_buffer == NULL) {
		dev_err(dev, "could not alloc %d byte inflate buffer\n",
			mcp_uncompressed_len);
		return -ENOMEM;
	}
	zs.workspace = vmalloc(zlib_inflate_workspacesize());
	if (zs.workspace == NULL) {
		dev_err(dev, "could not alloc zlib workspace\n");
		status = -ENOMEM;
		goto abort_with_inflate_buffer;
	}
	zerr = zlib_inflateInit(&zs);
	if (zerr != Z_OK) {
		dev_err(dev, "zlib_inflateInit fails with %d\n", zerr);
		goto abort_with_workspace;
	}
	zs.next_in = (char *)mcp;
	zs.avail_in = mcp_len;
	zs.next_out = (unsigned char *) inflate_buffer;
	zs.avail_out = mcp_uncompressed_len;
	zerr = zlib_inflate(&zs, Z_FINISH);
	if (zerr != Z_STREAM_END || zs.avail_in != 0 || zs.avail_out == 0) {
		dev_err(dev, "zlib_inflate fails with %d (%d %d)\n", 
			zerr, zs.avail_in, zs.avail_out);
		goto abort_with_workspace;
	}

	*size = zs.total_out;

	/* check id */
	hdr_offset = ntohl(*(__be32 *) (inflate_buffer + MCP_HEADER_PTR_OFFSET));
	if ((hdr_offset & 3) || hdr_offset + sizeof(*hdr) > *size) {
		dev_err(dev, "Bad firmware file\n");
		status = -EINVAL;
		goto abort_with_workspace;
	}
	hdr = (void*) (inflate_buffer + hdr_offset);

	status = myri10ge_validate_firmware(mgp, hdr);
	if (status != 0)
		goto abort_with_workspace;


	crc = crc32(~0, inflate_buffer, *size);
	for (i = 0; i < *size; i += 256) {
		myri10ge_pio_copy(mgp->sram + MYRI10GE_FW_OFFSET + i,
				  inflate_buffer + i,
				  min(256U, (unsigned)(*size - i)));
		mb();
		readb(mgp->sram);
		mb();
	}
	/* corruption checking is good for parity recovery and buggy chipset */
	memcpy_fromio(inflate_buffer, mgp->sram + MYRI10GE_FW_OFFSET, *size);
	reread_crc = crc32(~0, inflate_buffer, *size);
	if (crc != reread_crc) {
		dev_err(dev, "CRC failed(fw-len=%u), got 0x%x (expect 0x%x)\n",
		       (unsigned)*size, reread_crc, crc);
		status = -EIO;
		goto abort_with_workspace;
	}
	status = 0;

abort_with_workspace:
	vfree(zs.workspace);
abort_with_inflate_buffer:
	vfree(inflate_buffer);
	return status;
}

#else /* MYRI10GE_BUILTIN_FW */

#ifndef LINUX_KERNEL_SPECIFIC
#if !defined CONFIG_FW_LOADER && !defined CONFIG_FW_LOADER_MODULE
#error support for firmware_class (CONFIG_FW_LOADER=(y|m)) or MYRI10GE_BUILTIN_FW=1 required
#endif /* CONFIG_FW_LOADER */
#endif /* LINUX_KERNEL_SPECIFIC */

static int
myri10ge_load_hotplug_firmware(struct myri10ge_priv *mgp, u32 *size)
{
	unsigned crc, reread_crc;
	const struct firmware *fw;
	struct device *dev = &mgp->pdev->dev;
	struct mcp_gen_header *hdr;
	size_t hdr_offset;
	int status;
	unsigned i;

	if ((status = request_firmware(&fw, mgp->fw_name, dev)) < 0) {
		dev_err(dev, "Unable to load %s firmware image via hotplug\n",
			mgp->fw_name);
		status = -EINVAL;
		goto abort_with_nothing;
	}

	/* check size */

	if (fw->size >= mgp->sram_size - MYRI10GE_FW_OFFSET ||
	    fw->size < MCP_HEADER_PTR_OFFSET + 4) {
		dev_err(dev, "Firmware size invalid:%d\n", (int)fw->size);
		status = -EINVAL;
		goto abort_with_fw;
	}

	/* check id */
	hdr_offset = ntohl(*(__be32 *) (fw->data + MCP_HEADER_PTR_OFFSET));
	if ((hdr_offset & 3) || hdr_offset + sizeof(*hdr) > fw->size) {
		dev_err(dev, "Bad firmware file\n");
		status = -EINVAL;
		goto abort_with_fw;
	}
	hdr = (void*) (fw->data + hdr_offset);

	status = myri10ge_validate_firmware(mgp, hdr);
	if (status != 0)
		goto abort_with_fw;

	crc = crc32(~0, fw->data, fw->size);
	for (i = 0; i < fw->size; i += 256) {
		myri10ge_pio_copy(mgp->sram + MYRI10GE_FW_OFFSET + i,
				  fw->data + i,
				  min(256U, (unsigned)(fw->size - i)));
		mb();
		readb(mgp->sram);
#ifndef LINUX_KERNEL_SPECIFIC
		/* considered way too superfluous by the kernel guys */
		mb();
#endif
	}
	/* corruption checking is good for parity recovery and buggy chipset */
	memcpy_fromio(fw->data, mgp->sram + MYRI10GE_FW_OFFSET, fw->size);
	reread_crc = crc32(~0, fw->data, fw->size);
	if (crc != reread_crc) {
		dev_err(dev, "CRC failed(fw-len=%u), got 0x%x (expect 0x%x)\n",
		       (unsigned)fw->size, reread_crc, crc);
		status = -EIO;
		goto abort_with_fw;
	}
	*size = (u32)fw->size;

abort_with_fw:
	release_firmware(fw);

abort_with_nothing:
	return status;
}
#endif /* MYRI10GE_BUILTIN_FW */

static int
myri10ge_adopt_running_firmware(struct myri10ge_priv *mgp)
{
	struct mcp_gen_header *hdr;
	struct device *dev = &mgp->pdev->dev;
	const size_t bytes = sizeof (struct mcp_gen_header);
	size_t hdr_offset;
	int status;

	/* find running firmware header */
	hdr_offset = ntohl(__raw_readl(mgp->sram + MCP_HEADER_PTR_OFFSET));

	if ((hdr_offset & 3) || hdr_offset + sizeof(*hdr) > mgp->sram_size) {
		dev_err(dev, "Running firmware has bad header offset (%d)\n",
			(int)hdr_offset);
		return -EIO;
	}

	/* copy header of running firmware from SRAM to host memory to
	 * validate firmware */
	hdr = kmalloc(bytes, GFP_KERNEL);
	if (hdr == NULL) {
		dev_err(dev, "could not malloc firmware hdr\n");
		return -ENOMEM;
	}
	memcpy_fromio(hdr, mgp->sram + hdr_offset, bytes);
	status = myri10ge_validate_firmware(mgp, hdr);
	kfree(hdr);

	/* check to see if adopted firmware has bug where adopting
	 * it will cause broadcasts to be filtered unless the NIC
	 * is kept in ALLMULTI mode */
	if (mgp->fw_ver_major == 1 && mgp->fw_ver_minor == 4 &&
	    mgp->fw_ver_tiny >= 4 && mgp->fw_ver_tiny <= 11) {
		mgp->adopted_rx_filter_bug = 1;
		dev_warn(dev, "Adopting fw %d.%d.%d: "
			 "working around rx filter bug\n",
			 mgp->fw_ver_major, mgp->fw_ver_minor,
			 mgp->fw_ver_tiny);
	}
	return status;
}


static int
myri10ge_load_firmware(struct myri10ge_priv *mgp)
{
	char __iomem *submit;
	__be32 buf[16];
	u32 dma_low, dma_high, size;
	int status, i;

	size = 0;
#if MYRI10GE_BUILTIN_FW
	status = myri10ge_load_builtin_firmware(mgp, &size);
#else /* MYRI10GE_BUILTIN_FW */
	status = myri10ge_load_hotplug_firmware(mgp, &size);
#endif /* MYRI10GE_BUILTIN_FW */
	if (status) {
		dev_warn(&mgp->pdev->dev,
#if MYRI10GE_BUILTIN_FW
			 "firmware loading failed\n");
#else
			 "hotplug firmware loading failed\n");
#endif /* MYRI10GE_BUILTIN_FW */

		/* Do not attempt to adopt firmware if there
		   was a bad crc */
		if (status == -EIO)
			return status;

		status = myri10ge_adopt_running_firmware(mgp);
		if (status != 0) {
			dev_err(&mgp->pdev->dev,
				"failed to adopt running firmware\n");
			return status;
		}
		dev_info(&mgp->pdev->dev,
			 "Successfully adopted running firmware\n");
		if (mgp->tx.boundary == 4096) {
			dev_warn(&mgp->pdev->dev,
				"Using firmware currently running on NIC"
				 ".  For optimal\n");
			dev_warn(&mgp->pdev->dev,
				 "performance consider loading optimized "
				 "firmware\n");
			dev_warn(&mgp->pdev->dev, "via hotplug\n");
		}

		mgp->fw_name = "adopted";
		mgp->tx.boundary = 2048;
		return status;
	}

	/* clear confirmation addr */
	mgp->cmd->data = 0;
	mb();

	/* send a reload command to the bootstrap MCP, and wait for the
	 *  response in the confirmation address.  The firmware should
	 * write a -1 there to indicate it is alive and well
	 */
	dma_low = MYRI10GE_LOWPART_TO_U32(mgp->cmd_bus);
	dma_high = MYRI10GE_HIGHPART_TO_U32(mgp->cmd_bus);

	buf[0] = htonl(dma_high); 	/* confirm addr MSW */
	buf[1] = htonl(dma_low); 	/* confirm addr LSW */
	buf[2] = MYRI10GE_NO_CONFIRM_DATA;	/* confirm data */

	/* FIX: All newest firmware should un-protect the bottom of
	 * the sram before handoff. However, the very first interfaces
	 * do not. Therefore the handoff copy must skip the first 8 bytes
	 */
	buf[3] = htonl(MYRI10GE_FW_OFFSET + 8);	/* where the code starts */
	buf[4] = htonl(size - 8); 		/* length of code */
	buf[5] = htonl(8);			/* where to copy to */
	buf[6] = htonl(0);			/* where to jump to */

	submit = mgp->sram + MXGEFW_BOOT_HANDOFF;

	myri10ge_pio_copy(submit, &buf, sizeof (buf));
	mb();
	myri10ge_msleep(1);
	mb();
	i = 0;
	while (mgp->cmd->data != MYRI10GE_NO_CONFIRM_DATA && i < 20) {
		myri10ge_msleep(1);
		i++;
	}
	if (mgp->cmd->data != MYRI10GE_NO_CONFIRM_DATA) {
		dev_err(&mgp->pdev->dev, "handoff failed\n");
		return -ENXIO;
	}
	dev_info(&mgp->pdev->dev, "handoff confirmed\n");
	myri10ge_dummy_rdma(mgp, 1);

	return 0;
}

static int
myri10ge_update_mac_address(struct myri10ge_priv *mgp, u8 *addr)
{
	struct myri10ge_cmd cmd;
	int status;

	cmd.data0 = ((addr[0] << 24) | (addr[1] << 16)
		     | (addr[2] << 8) | addr[3]);

	cmd.data1 = ((addr[4] << 8) | (addr[5]));

	status = myri10ge_send_cmd(mgp, MXGEFW_SET_MAC_ADDRESS, &cmd, 0);
	return status;
}

static int
myri10ge_change_pause(struct myri10ge_priv *mgp, int pause)
{
	struct myri10ge_cmd cmd;
	int status, ctl;

	ctl = pause ? MXGEFW_ENABLE_FLOW_CONTROL :
		MXGEFW_DISABLE_FLOW_CONTROL;
	status = myri10ge_send_cmd(mgp, ctl, &cmd, 0);

	if (status) {
		printk(KERN_ERR "myri10ge: %s: Failed to set flow control mode\n",
		       mgp->dev->name);
		return status;
	}
	mgp->pause = pause;
	return 0;
}

static void
myri10ge_change_promisc(struct myri10ge_priv *mgp, int promisc, int atomic)
{
	struct myri10ge_cmd cmd;
	int status, ctl;

	ctl = promisc ? MXGEFW_ENABLE_PROMISC :
		MXGEFW_DISABLE_PROMISC;
	status = myri10ge_send_cmd(mgp, ctl, &cmd, atomic);
	if (status)
		printk(KERN_ERR "myri10ge: %s: Failed to set promisc mode\n",
		       mgp->dev->name);
}

static int
myri10ge_dma_test(struct myri10ge_priv *mgp, int test_type)
{
	struct myri10ge_cmd cmd;
	int status;
	u32 len;
	struct page *dmatest_page;
	dma_addr_t dmatest_bus;
	char *test = " ";

	dmatest_page = alloc_page(GFP_KERNEL);
	if (!dmatest_page)
		return -ENOMEM;
	dmatest_bus = pci_map_page(mgp->pdev, dmatest_page, 0, PAGE_SIZE, 
				   DMA_BIDIRECTIONAL);

	/* Run a small DMA test.
	 * The magic multipliers to the length tell the firmware
	 * to do DMA read, write, or read+write tests.  The
	 * results are returned in cmd.data0.  The upper 16
	 * bits or the return is the number of transfers completed.
	 * The lower 16 bits is the time in 0.5us ticks that the
	 * transfers took to complete.
	 */

	len = mgp->tx.boundary;

	cmd.data0 = MYRI10GE_LOWPART_TO_U32(dmatest_bus);
	cmd.data1 = MYRI10GE_HIGHPART_TO_U32(dmatest_bus);
	cmd.data2 = len * 0x10000;
	status = myri10ge_send_cmd(mgp, test_type, &cmd, 0);
	if (status != 0) {
		test = "read";
		goto abort;
	}
	mgp->read_dma = ((cmd.data0>>16) * len * 2) /
		(cmd.data0 & 0xffff);
	cmd.data0 = MYRI10GE_LOWPART_TO_U32(dmatest_bus);
	cmd.data1 = MYRI10GE_HIGHPART_TO_U32(dmatest_bus);
	cmd.data2 = len * 0x1;
	status = myri10ge_send_cmd(mgp, test_type, &cmd, 0);
	if (status != 0) {
		test = "write";
		goto abort;
	}
	mgp->write_dma = ((cmd.data0>>16) * len * 2) /
		(cmd.data0 & 0xffff);

	cmd.data0 = MYRI10GE_LOWPART_TO_U32(dmatest_bus);
	cmd.data1 = MYRI10GE_HIGHPART_TO_U32(dmatest_bus);
	cmd.data2 = len * 0x10001;
	status = myri10ge_send_cmd(mgp, test_type, &cmd, 0);
	if (status != 0) {
		test = "read/write";
		goto abort;
	}
	mgp->read_write_dma = ((cmd.data0>>16) * len * 2 * 2) /
		(cmd.data0 & 0xffff);

abort:
	pci_unmap_page(mgp->pdev, dmatest_bus, PAGE_SIZE, DMA_BIDIRECTIONAL);
	put_page(dmatest_page);

	if (status != 0 && test_type != MXGEFW_CMD_UNALIGNED_TEST)
		dev_warn(&mgp->pdev->dev, "DMA %s benchmark failed: %d\n",
			 test, status);

	return status;
}

static int
myri10ge_reset(struct myri10ge_priv *mgp)
{
	struct myri10ge_cmd cmd;
	int status;
	size_t bytes;

	/* try to send a reset command to the card to see if it
	   is alive */
	memset(&cmd, 0, sizeof (cmd));
	status = myri10ge_send_cmd(mgp, MXGEFW_CMD_RESET, &cmd, 0);
	if (status != 0) {
		dev_err(&mgp->pdev->dev, "failed reset\n");
		return -ENXIO;
	}

	(void)myri10ge_dma_test(mgp, MXGEFW_DMA_TEST);

	/* Now exchange information about interrupts  */

	bytes = myri10ge_max_intr_slots * sizeof (*mgp->rx_done.entry);
	memset(mgp->rx_done.entry, 0, bytes);
	cmd.data0 = (u32) bytes;
	status = myri10ge_send_cmd(mgp, MXGEFW_CMD_SET_INTRQ_SIZE, &cmd, 0);
	cmd.data0 = MYRI10GE_LOWPART_TO_U32(mgp->rx_done.bus);
	cmd.data1 = MYRI10GE_HIGHPART_TO_U32(mgp->rx_done.bus);
	status |= myri10ge_send_cmd(mgp, MXGEFW_CMD_SET_INTRQ_DMA, &cmd, 0);


	status |= myri10ge_send_cmd(mgp,  MXGEFW_CMD_GET_IRQ_ACK_OFFSET, &cmd, 0);
	mgp->irq_claim = (__iomem __be32 *) (mgp->sram + cmd.data0);
	status |= myri10ge_send_cmd(mgp,  MXGEFW_CMD_GET_IRQ_DEASSERT_OFFSET,
				    &cmd, 0);
	mgp->irq_deassert = (__iomem __be32 *) (mgp->sram + cmd.data0);

	status |= myri10ge_send_cmd
		(mgp, MXGEFW_CMD_GET_INTR_COAL_DELAY_OFFSET, &cmd, 0);
	mgp->intr_coal_delay_ptr = (__iomem __be32 *) (mgp->sram + cmd.data0);
	if (status != 0) {
		dev_err(&mgp->pdev->dev, "failed set interrupt parameters\n");
		return status;
	}
	put_be32(htonl(mgp->intr_coal_delay), mgp->intr_coal_delay_ptr);

	memset(mgp->rx_done.entry, 0, bytes);

	/* reset mcp/driver shared state back to 0 */
	mgp->tx.req = 0;
	mgp->tx.done = 0;
	mgp->tx.pkt_start = 0;
	mgp->tx.pkt_done = 0;
	mgp->rx_big.cnt = 0;
	mgp->rx_small.cnt = 0;
	mgp->rx_done.idx = 0;
	mgp->rx_done.cnt = 0;
	mgp->link_changes = 0;
#if MYRI10GE_LRO
	mgp->lro_queued = 0;
	mgp->lro_flushed = 0;
	mgp->lro_too_many_streams = 0;
#endif /* MYRI10GE_LRO */
	status = myri10ge_update_mac_address(mgp, mgp->dev->dev_addr);
	myri10ge_change_pause(mgp, mgp->pause);
	myri10ge_set_multicast_list(mgp->dev);
	return status;
}

static inline void
myri10ge_submit_8rx(struct mcp_kreq_ether_recv __iomem *dst, struct mcp_kreq_ether_recv *src)
{
	__be32 low;

	low = src->addr_low;
	src->addr_low = htonl(DMA_32BIT_MASK);
	myri10ge_pio_copy(dst, src, 4 * sizeof(*src));
	mb();
	myri10ge_pio_copy(dst + 4, src + 4, 4 * sizeof(*src));
	mb();
	src->addr_low = low;
	put_be32(low, &dst->addr_low);
	mb();
}



static inline void
myri10ge_vlan_ip_csum(struct sk_buff *skb, __wsum hw_csum)
{
	struct vlan_hdr *vh = (struct vlan_hdr *) (skb->data);

	if ((skb->protocol == htons(ETH_P_8021Q)) &&
	    (vh->h_vlan_encapsulated_proto == htons(ETH_P_IP) ||
	     vh->h_vlan_encapsulated_proto == htons(ETH_P_IPV6))) {
		skb->csum = hw_csum;
		skb->ip_summed = CHECKSUM_COMPLETE;
#ifndef LINUX_KERNEL_SPECIFIC
		if (myri10ge_vlan_csum_fixup)
			skb->csum =
				csum_sub(skb->csum,
					 csum_partial(skb->data,
						      VLAN_HLEN, 0));
#endif /* LINUX_KERNEL_SPECIFIC */
	}
}

static inline void
myri10ge_rx_skb_build(struct sk_buff *skb, u8 *va, struct skb_frag_struct *rx_frags, 
		      int len, int hlen)
{
	struct skb_frag_struct *skb_frags;

	skb->len = skb->data_len = len;
	skb->truesize = len + sizeof (struct sk_buff);
	/* attach the page(s) */

	skb_frags = skb_shinfo(skb)->frags;
	while (len > 0) {
		memcpy(skb_frags, rx_frags, sizeof (*skb_frags));
		len -= rx_frags->size;
		skb_frags++;
		rx_frags++;
		skb_shinfo(skb)->nr_frags++;
	}

	/* pskb_may_pull is not available in irq context, but
	   skb_pull() (for ether_pad and eth_type_trans()) requires
	   the beginning of the packet in skb_headlen(), move it
	   manually */
	myri10ge_skb_copy_to_linear_data(skb, va, hlen);
	skb_shinfo(skb)->frags[0].page_offset += hlen;
	skb_shinfo(skb)->frags[0].size -= hlen;
	skb->data_len -= hlen;
	skb->tail += hlen;
	skb_pull(skb, MXGEFW_PAD);
}

#if MYRI10GE_LRO

/* debug aid to check for "bad" hardware */
#define MYRI10GE_LRO_CSUM_UNNECESSARY 1
#define MYRI10GE_LRO_DISABLE_HW_CSUM 0

static void
myri10ge_frag_trim(struct skb_frag_struct *rx_frags,
		   int old_len, int trim)
{
	struct skb_frag_struct *frag;
	int offset = 0;
	int new_len = old_len - trim;
	int old_size;

	/* find the frag where the IP payload ends. This
	   should almost always be the 1st fragment */
	frag = rx_frags;
	while (offset + frag->size < new_len) {
		offset += frag->size;
		frag++;
	}
	/* adjust its length */
	old_size = frag->size;
	frag->size = new_len - offset;
	
	/* release any excess pages */
	offset += old_size;
	while (offset < old_len) {
		frag++;
		offset +=  frag->size;
		put_page(frag->page);
	}
}

static inline int
myri10ge_lro_csum(int tcplen, struct iphdr *iph, __wsum csum)
{
	if (unlikely(ip_fast_csum((u8 *)iph, iph->ihl)))
		return -1;

	if (unlikely(csum_tcpudp_magic(iph->saddr, iph->daddr,
				       tcplen, IPPROTO_TCP, csum)))
		return -1;
	return 0;
}

static inline void
myri10ge_lro_flush(struct myri10ge_priv *mgp, struct myri10ge_lro_packet *lro)
{
	struct iphdr *iph;
	struct tcphdr *th;
	struct sk_buff *skb;
	u32 *ts_ptr;
	u32 tcplen;

	skb = lro->skb;

	if (lro->append_cnt) {
		/* incorporate the new len into the ip header and
		 * re-calculate the checksum,  Note that 
		 * eth_type_trans() left skb->data at the start of
		 * the vlan header, so we need to skip past it to
		 * get to the IP header */
		if (lro->vlan_tci) {
			iph = (struct iphdr *) (skb->data + VLAN_HLEN);
			iph->tot_len = htons(skb->len - VLAN_HLEN);
		} else {
			iph = (struct iphdr *)skb->data;
			iph->tot_len = htons(skb->len);
		}
		iph->check = 0;
		iph->check = ip_fast_csum((unsigned char *)iph, iph->ihl);

		/* incorporate the latest ack into the tcp header */
		th = (struct tcphdr *) (iph + 1);
		th->ack_seq = lro->ack_seq;
		th->window = lro->window;

		/* incorporate latest timestamp into the tcp header */
		if (lro->timestamp) {
			ts_ptr = (u32 *)(th + 1);
			ts_ptr[1] = htonl(lro->tsval);
			ts_ptr[2] = lro->tsecr;
		}

		/* 
		 * update checksum in tcp header by re-calculating the
		 * tcp pseudoheader checksum, and adding it to the checksum
		 * of the tcp payload data 
		 */
		th->check = 0;
		tcplen = ntohs(iph->tot_len) - sizeof(*iph);
		th->check = myri10ge_tcp_v4_check(th, tcplen, iph->saddr, iph->daddr,
						  csum_partial((char *)th,
							       /* TODO: use myri10ge_tcp_hdrlen? */
							       th->doff << 2,
							       lro->data_csum));
#if !MYRI10GE_LRO_CSUM_UNNECESSARY

		/* finally, reconstruct a "hardware"  checksum by
		 * adding the tcp header checksum to the tcp data
		 * checksum
		 */
		skb->csum =  csum_partial((char *)th,
					  /* TODO: use myri10ge_tcp_hdrlen? */
					  th->doff << 2,
					  lro->data_csum);
#endif				     

		skb->truesize = skb->len + sizeof(struct sk_buff);
	}
#if !MYRI10GE_LRO_CSUM_UNNECESSARY
	if (lro->vlan_tci && !myri10ge_vlan_csum_fixup) {
		/* we subtracted out the VLAN checksum to check the
		   checksum in vlan_lro_rx().  It must be replaced
		   prior to passing it up the stack */
		skb->csum = csum_partial(skb->data,
					 VLAN_HLEN, skb->csum);
	} 
#if MYRI10GE_LRO_DISABLE_HW_CSUM
	skb->csum = 0;
	skb->ip_summed = CHECKSUM_NONE;
#endif
#endif

	skb_shinfo(skb)->MYRI10GE_GSO_SIZE = lro->mss;
#ifdef MYRI10GE_NAPI
	netif_receive_skb(skb);
#else
	netif_rx(skb);
#endif
	mgp->dev->last_rx = jiffies;
	mgp->lro_queued += lro->append_cnt + 1;
	mgp->lro_flushed++;
	lro->skb = NULL;
	lro->timestamp = 0;
	lro->append_cnt = 0;
	hlist_add_head(&lro->lro_node, &mgp->rx_done.lro_free);
}

static int
myri10ge_lro_rx(struct myri10ge_priv *mgp, u8 *va,
		struct skb_frag_struct *rx_frags, int *len, __wsum csum)
{
	struct ethhdr *eh;
	struct vlan_ethhdr *vh;
	struct iphdr *iph;
	struct tcphdr *th;
	struct myri10ge_lro_packet *lro;
	u32 *ts_ptr = NULL; /* XXX -Wuninitialized */
	struct sk_buff *skb;
	struct skb_frag_struct *skb_frags;
	struct hlist_node *node;
	int opt_bytes, tcp_data_len, tcp_hdr_len, hlen, trim, llhlen;
	__u32 seq;
	__u16 ip_len, vlan_tci;

	/* check to see that it is IP */
	eh = (struct ethhdr *) (va + MXGEFW_PAD);
	if (eh->h_proto == ntohs(ETH_P_IP)) {
		llhlen = ETH_HLEN;
		vlan_tci = 0;
	} else if (eh->h_proto == ntohs(ETH_P_8021Q)) {
		vh = (struct vlan_ethhdr *) (va + MXGEFW_PAD);
		if (vh->h_vlan_encapsulated_proto !=  ntohs(ETH_P_IP))
			return -1;
		llhlen = VLAN_ETH_HLEN;
		vlan_tci = vh->h_vlan_TCI;
		/* HW checksum starts after the ethernet header, we
		   must subtract off the VLAN header's checksum before
		   csum can be used */
		csum = csum_sub(csum, 
				csum_partial(va + MXGEFW_PAD + ETH_HLEN,
					     VLAN_HLEN, 0));
	} else {
		return -1;
	}

	/* now check to see if it is TCP */
	iph = (struct iphdr *)(va + llhlen + MXGEFW_PAD);
	if (iph->protocol != IPPROTO_TCP)
		return -1;

	/* ensure there are no options */
	if ((iph->ihl << 2) != sizeof (*iph))
		return -1;

	/* .. and the packet is not fragmented */
	if (iph->frag_off & htons(IP_MF|IP_OFFSET))
		return -1;

	/* find the TCP header */
	th = (struct tcphdr *) (iph + 1);
	
	/* ensure no bits set besides ack or psh */
	if (th->fin || th->syn || th->rst || th->urg || th->ece 
	    || th->cwr || !th->ack)
		return -1;

	/* check for timestamps. Since the only option we handle are
	   timestamps, we only have to handle the simple case of
	   aligned timestamps */

	/* TODO: use myri10ge_tcp_hdrlen? */
	opt_bytes = (th->doff << 2) - sizeof (*th);
	tcp_hdr_len =  sizeof (*th) + opt_bytes;
	if (opt_bytes != 0) {
		ts_ptr = (u32 *)(th + 1);
		if (unlikely(opt_bytes != TCPOLEN_TSTAMP_ALIGNED) ||
		    (*ts_ptr !=  ntohl((TCPOPT_NOP << 24) | (TCPOPT_NOP << 16)
					| (TCPOPT_TIMESTAMP << 8) 
					| TCPOLEN_TIMESTAMP))) {
			return -1;
		}
	}

	ip_len = ntohs(iph->tot_len);
	/* TODO: use myri10ge_tcp_hdrlen? */
	tcp_data_len = ip_len - (th->doff << 2) - sizeof (*iph);

	if (unlikely(myri10ge_lro_csum(tcp_hdr_len + tcp_data_len, iph, csum))) {
		mgp->lro_bad_csum++;		
		return -1;
	}


	/* 
	 * If frame is padded beyond the end of the IP packet,
	 * then we must trim the extra bytes off the end.
	 */
	trim = *len - (ip_len + llhlen + MXGEFW_PAD);
	if (trim != 0) {
		/* ensure we received the full frame */
		if (unlikely(trim < 0))
			return -1;
		/* trim off any padding */
		myri10ge_frag_trim(rx_frags, *len, trim);
		*len -= trim;
	}

	hlen = ip_len + llhlen - tcp_data_len;

	seq = ntohl(th->seq);

	/* now we have a packet that might be eligible for LRO,
	   so see if it matches anything we might expect */

	hlist_for_each_entry(lro, node, &mgp->rx_done.lro_active, lro_node) {
		if (lro->source_port == th->source && 
		    lro->dest_port == th->dest &&
		    lro->source_ip == iph->saddr && 
		    lro->dest_ip == iph->daddr && 
		    lro->vlan_tci == vlan_tci) {
			/* Try to append it */

			if (unlikely(seq != lro->next_seq)) {
				/* out of order packet */
				hlist_del(&lro->lro_node);
				myri10ge_lro_flush(mgp, lro);
				return -1;
			}
			if (opt_bytes) {
				__u32 tsval = ntohl(*(ts_ptr + 1));
				/* make sure timestamp values are increasing */
				if (unlikely(lro->tsval > tsval || 
					     *(ts_ptr + 2) == 0)) {
					return -1;
				}
				lro->tsval = tsval;
				lro->tsecr = *(ts_ptr + 2);
			}
			lro->next_seq += tcp_data_len;
			lro->ack_seq = th->ack_seq;
			skb = lro->skb;

			/* subtract off the checksum of the tcp header
			 * from the hardware checksum, and add it to the
			 * stored tcp data checksum.  csum_block_add()
			 * is used, as the total length so far may be
			 * odd 
			 */
			lro->data_csum = 
				csum_block_add(lro->data_csum,
					       csum_sub(csum,
							csum_partial((u8 *)th, 
								     tcp_hdr_len,
								     0)),
					       skb->data_len);
			lro->window = th->window;
			skb->data_len += tcp_data_len;
			skb->len += tcp_data_len;
			if (tcp_data_len > lro->mss)
				lro->mss = tcp_data_len;

			/* pull off the header and firmware pad
			   before we copy the data */

			hlen += MXGEFW_PAD;
			rx_frags[0].page_offset += hlen;
			rx_frags[0].size -= hlen;
			*len -= hlen;
			skb_frags = &skb_shinfo(skb)->frags[skb_shinfo(skb)->nr_frags];
			/* if it was just header (like a TCP ack with
			   no data), release the page */
			if (*len <= 0) {
				put_page(rx_frags[0].page);
			} else {
				while (*len > 0) {
					memcpy(skb_frags, rx_frags,
					       sizeof (*skb_frags));
					*len -= rx_frags->size;
					rx_frags++;
					skb_frags++;
					skb_shinfo(skb)->nr_frags++;
				}
			}

			lro->append_cnt++;

			/* cheap, conservative test.  We may waste
			 * some slots with a 1500 byte mtu */
			if (skb_shinfo(skb)->nr_frags 
			    + MYRI10GE_MAX_FRAGS_PER_FRAME > MAX_SKB_FRAGS 
			    || mgp->dev->mtu + skb->len > 65535) {
				hlist_del(&lro->lro_node);
				myri10ge_lro_flush(mgp, lro);
			}
			return 0;
		}
	}
		
	/* start a new packet */
	if (!hlist_empty(&mgp->rx_done.lro_free)) {
		lro = hlist_entry(mgp->rx_done.lro_free.first, 
				  struct myri10ge_lro_packet, lro_node);
		/* allocate an skb to attach the page(s) to */

		skb = myri10ge_netdev_alloc_skb(mgp->dev,  hlen + 16);
		if (unlikely(skb == NULL))
			return -1;

		myri10ge_rx_skb_build(skb, va, rx_frags, *len, hlen + MXGEFW_PAD);
		skb->protocol = eth_type_trans(skb, mgp->dev);
#ifndef LINUX_KERNEL_SPECIFIC
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,21)
		/* eth_type_trans sets skb->dev in 2.6.22 */
		skb->dev = mgp->dev;
#endif
#endif
#if MYRI10GE_LRO_CSUM_UNNECESSARY
		skb->ip_summed = CHECKSUM_UNNECESSARY;
#else
		skb->csum = csum;
		skb->ip_summed = CHECKSUM_COMPLETE;
#endif
		lro->skb = skb;
		lro->source_ip = iph->saddr;
		lro->dest_ip = iph->daddr;
		lro->source_port = th->source;
		lro->dest_port = th->dest;
		lro->next_seq = seq + tcp_data_len;
		lro->mss = tcp_data_len;
		lro->ack_seq = th->ack_seq;

		/* save the checksum of just the TCP payload by
		 * subtracting off the checksum of the TCP header from
		 * the entire hardware checksum 
		 */
		lro->data_csum = csum_sub(csum, 
					  csum_partial((u8 *)th,
						       tcp_hdr_len, 0));
		lro->window = th->window;
		lro->vlan_tci = vlan_tci;
		/* record timestamp if it is present */
		if (opt_bytes) {
			lro->timestamp = 1;
			lro->tsval = ntohl(*(ts_ptr + 1));
			lro->tsecr = *(ts_ptr + 2);
		}
		/* remove first packet from freelist.. */
		hlist_del(&lro->lro_node);
		/* .. and insert at the front of the active list */
		hlist_add_head(&lro->lro_node, &mgp->rx_done.lro_active);

		/* release the page if there was no data.  We do it
		 * down here since the code above refers to the
		 * contents of the page */
		if (skb_shinfo(skb)->frags[0].size <= 0) {
			put_page(skb_shinfo(skb)->frags[0].page);
			skb_shinfo(skb)->nr_frags = 0;
		}
		return 0;
	}
	mgp->lro_too_many_streams++;
	return -1;
}
#endif /* MYRI10GE_LRO */

#if MYRI10GE_RX_SKBS

#include "myri10ge_rx_skbs.h"

#else /* MYRI10GE_RX_SKBS */
static void
myri10ge_alloc_rx_pages(struct myri10ge_priv *mgp, struct myri10ge_rx_buf *rx,
			int bytes, int watchdog)
{
	struct page *page;
	int idx;


	if (unlikely(rx->watchdog_needed && !watchdog))
		return;

	/* try to refill entire ring */
	while (rx->fill_cnt != (rx->cnt + rx->mask + 1)) {
		idx = rx->fill_cnt & rx->mask;
		if (rx->page_offset + bytes <= MYRI10GE_ALLOC_SIZE) {
			/* we can use part of previous page */
			get_page(rx->page);
		} else {
			/* we need a new page */
			page = alloc_pages(GFP_ATOMIC | __GFP_COMP, MYRI10GE_ALLOC_ORDER);
			if (unlikely(page == NULL)) {
				if (rx->fill_cnt - rx->cnt < 16)
					rx->watchdog_needed = 1;
				return;
			}
			rx->page = page;
			rx->page_offset = 0;
			rx->bus = pci_map_page(mgp->pdev, page, 0,
					       MYRI10GE_ALLOC_SIZE, PCI_DMA_FROMDEVICE);
		}
		rx->info[idx].page = rx->page;
		rx->info[idx].page_offset = rx->page_offset;
		/* note that this is the address of the start of the
		 * page */
		pci_unmap_addr_set(&rx->info[idx], bus, rx->bus);
		rx->shadow[idx].addr_low = 
			htonl(MYRI10GE_LOWPART_TO_U32(rx->bus) + rx->page_offset);
		rx->shadow[idx].addr_high = 
			htonl(MYRI10GE_HIGHPART_TO_U32(rx->bus));

		/* start next packet on a cacheline boundary */
		rx->page_offset += SKB_DATA_ALIGN(bytes);

#if MYRI10GE_ALLOC_SIZE > 4096
		/* don't cross a 4KB boundary */
		if ((rx->page_offset >> 12) !=
		    ((rx->page_offset + bytes - 1) >> 12))
			rx->page_offset = (rx->page_offset + 4096) & ~4095;
#endif
		rx->fill_cnt++;

		/* copy 8 descriptors to the firmware at a time */
		if ((idx & 7) == 7) {
			if (rx->wc_fifo == NULL)
				myri10ge_submit_8rx(&rx->lanai[idx - 7],
						    &rx->shadow[idx - 7]);
			else {
				mb();
				myri10ge_pio_copy(rx->wc_fifo,
						  &rx->shadow[idx - 7], 64);
			}
		}
	}
}

static inline void
myri10ge_unmap_rx_page(struct pci_dev *pdev, 
		       struct myri10ge_rx_buffer_state *info, int bytes)
{
	/* unmap the recvd page if we're the only or last user of it */
	if (bytes >= MYRI10GE_ALLOC_SIZE/2 || 
	    (info->page_offset + 2 * bytes) > MYRI10GE_ALLOC_SIZE) {
		pci_unmap_page(pdev, 
			       (pci_unmap_addr(info, bus)
				& ~(MYRI10GE_ALLOC_SIZE - 1)), MYRI10GE_ALLOC_SIZE,
			       PCI_DMA_FROMDEVICE);
	}
}

#define MYRI10GE_HLEN 64	/* The number of bytes to copy from a
				 * page into an skb */

static inline int
myri10ge_rx_done(struct myri10ge_priv *mgp, struct myri10ge_rx_buf *rx,
                 int bytes, int len, __wsum csum)
{
        struct sk_buff *skb;
	struct skb_frag_struct rx_frags[MYRI10GE_MAX_FRAGS_PER_FRAME];
        int i, idx, hlen, remainder;
	struct pci_dev *pdev = mgp->pdev;
	struct net_device *dev = mgp->dev;
	u8 *va;

	len += MXGEFW_PAD;
	idx = rx->cnt & rx->mask;
	va = page_address(rx->info[idx].page) + rx->info[idx].page_offset;
	prefetch(va);
	/* Fill skb_frag_struct(s) with data from our receive */
	for (i = 0, remainder = len; remainder > 0;  i++) {
		myri10ge_unmap_rx_page(pdev, &rx->info[idx], bytes);
		rx_frags[i].page = rx->info[idx].page;
		rx_frags[i].page_offset = rx->info[idx].page_offset;
		if (remainder < MYRI10GE_ALLOC_SIZE)
			rx_frags[i].size = remainder;
		else
			rx_frags[i].size = MYRI10GE_ALLOC_SIZE;
		rx->cnt++;
		idx = rx->cnt & rx->mask;
		remainder -= MYRI10GE_ALLOC_SIZE;
	}

#if MYRI10GE_LRO
        if (mgp->csum_flag && myri10ge_lro &&
	    (0 == myri10ge_lro_rx(mgp, va, rx_frags, &len, csum)))
		return 1;
#endif
	hlen = MYRI10GE_HLEN > len ? len: MYRI10GE_HLEN;


	/* allocate an skb to attach the page(s) to. */
#if MYRI10GE_LRO
	/* This is done 
	   after trying LRO, so as to avoid skb allocation overheads */
#endif

	skb = myri10ge_netdev_alloc_skb(dev, MYRI10GE_HLEN + 16);
	if (unlikely(skb == NULL)) {
		mgp->stats.rx_dropped++;
		do {
			i--;
			put_page(rx_frags[i].page);
		} while (i != 0);
		return 0;
	}
	
	/* Attach the pages to the skb, and trim off any padding */
	myri10ge_rx_skb_build(skb, va, rx_frags, len, hlen);
	if (skb_shinfo(skb)->frags[0].size <= 0) {
		put_page(skb_shinfo(skb)->frags[0].page);
		skb_shinfo(skb)->nr_frags = 0;
	}
	skb->protocol = eth_type_trans(skb, dev);
#ifndef LINUX_KERNEL_SPECIFIC
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,21)
	/* eth_type_trans sets skb->dev in 2.6.22 */
	skb->dev = dev;
#endif

	/* 
	 *  Old kernels have bugs which can cause problems dealing
	 *  with non-linear frames in oddball codepaths.  To work
	 *  around this, we linearize everything but IP and IPv6.
	 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)

/* ensure that NAPI is enabled so that we are not called in an irq
 * context, and skb_linearize() will work on HIGHMEM kernels
 */
#if !defined(MYRI10GE_NAPI) && defined(CONFIG_HIGHMEM)
#error You must enable MYRI10GE_NAPI on HIGHMEM kernels < 2.6.18 so that skb_linearize can be used safely
#endif /* !defined(MYRI10GE_NAPI) && defined(CONFIG_HIGHMEM) */

	{
		struct vlan_hdr *vh = (struct vlan_hdr *) (skb->data);

		/* linearize anything but IP or IPv6 */
		if (skb->protocol != htons(ETH_P_IP) &&
		    skb->protocol != htons(ETH_P_IPV6) &&
		    !(skb->protocol == htons(ETH_P_8021Q) &&
			    (vh->h_vlan_encapsulated_proto == htons(ETH_P_IP) ||
			     vh->h_vlan_encapsulated_proto == htons(ETH_P_IPV6)))) {
			if (myri10ge_skb_linearize(skb)) {
				dev_kfree_skb_any(skb);
				mgp->stats.rx_dropped++;
				return 0;
			}
		}
	}
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18) */
#endif /* LINUX_KERNEL_SPECIFIC */

        if (mgp->csum_flag) {
		if ((skb->protocol == htons(ETH_P_IP)) ||
		    (skb->protocol == htons(ETH_P_IPV6))) {
			skb->csum = csum;
			skb->ip_summed = CHECKSUM_COMPLETE;
		} else
			myri10ge_vlan_ip_csum(skb, csum);
        }
#ifdef MYRI10GE_NAPI
	netif_receive_skb(skb);
#else
	netif_rx(skb);
#endif
	dev->last_rx = jiffies;
	return 1;
}

#endif /* MYRI10GE_RX_SKBS */

static inline void
myri10ge_tx_done(struct myri10ge_priv *mgp, int mcp_index)
{
	struct pci_dev *pdev = mgp->pdev;
	struct myri10ge_tx_buf *tx = &mgp->tx;
	struct sk_buff *skb;
	int idx, len;

	while (tx->pkt_done != mcp_index) {
		idx = tx->done & tx->mask;
		skb = tx->info[idx].skb;

		/* Mark as free */
		tx->info[idx].skb = NULL;
		if (tx->info[idx].last) {
			tx->pkt_done++;
			tx->info[idx].last = 0;
		}
		tx->done++;
		len = pci_unmap_len(&tx->info[idx], len);
		pci_unmap_len_set(&tx->info[idx], len, 0);
		if (skb) {
			mgp->stats.tx_bytes += skb->len;
			mgp->stats.tx_packets++;
			dev_kfree_skb_irq(skb);
			if (len)
				pci_unmap_single(pdev,
						 pci_unmap_addr(&tx->info[idx], bus),
						 len, PCI_DMA_TODEVICE);
		} else {
			if (len)
				pci_unmap_page(pdev,
					       pci_unmap_addr(&tx->info[idx], bus),
					       len, PCI_DMA_TODEVICE);
		}
	}
	/* start the queue if we've stopped it */
	if (netif_queue_stopped(mgp->dev)
	    && tx->req - tx->done < (tx->mask >> 1)) {
		mgp->wake_queue++;
		netif_wake_queue(mgp->dev);
	}
}

static inline void
myri10ge_clean_rx_done(struct myri10ge_priv *mgp, int *limit)
{
	struct myri10ge_rx_done *rx_done = &mgp->rx_done;
#if MYRI10GE_LRO
	struct hlist_node *node, *node2;
	struct myri10ge_lro_packet *lro;
#endif
	unsigned long rx_bytes = 0;
	unsigned long rx_packets = 0;
	unsigned long rx_ok;

	int idx = rx_done->idx;
	int cnt = rx_done->cnt;
	u16 length;
	__wsum checksum;

	while (rx_done->entry[idx].length != 0 &&
		*limit != 0) {
		length = ntohs(rx_done->entry[idx].length);
		rx_done->entry[idx].length = 0;
		checksum = csum_unfold(rx_done->entry[idx].checksum);
		if (length <= mgp->small_bytes)
			rx_ok = myri10ge_rx_done(mgp, &mgp->rx_small,
						 mgp->small_bytes,
						 length, checksum);
		else
			rx_ok = myri10ge_rx_done(mgp, &mgp->rx_big,
						 mgp->big_bytes,
						 length, checksum);
		rx_packets += rx_ok;
		rx_bytes += rx_ok * (unsigned long)length;
		cnt++;
		idx = cnt & (myri10ge_max_intr_slots - 1);

		/* limit potential for livelock by only handling a
		 * limited number of frames. */
		(*limit)--;
	}
	rx_done->idx = idx;
	rx_done->cnt = cnt;
#if MYRI10GE_LRO
	hlist_for_each_entry_safe(lro, node, node2, &mgp->rx_done.lro_active, lro_node) {
		hlist_del(&lro->lro_node);
		myri10ge_lro_flush(mgp, lro);
	}
#endif /* MYRI10GE_LRO */
	mgp->stats.rx_packets += rx_packets;
	mgp->stats.rx_bytes += rx_bytes;

#if !MYRI10GE_RX_SKBS
	/* restock receive rings if needed */
	if (mgp->rx_small.fill_cnt - mgp->rx_small.cnt < myri10ge_fill_thresh)
		myri10ge_alloc_rx_pages(mgp, &mgp->rx_small,
					mgp->small_bytes + MXGEFW_PAD, 0);
	if (mgp->rx_big.fill_cnt - mgp->rx_big.cnt < myri10ge_fill_thresh)
		myri10ge_alloc_rx_pages(mgp, &mgp->rx_big, mgp->big_bytes, 0);

#endif /* MYRI10GE_RX_SKBS */
}

static inline void
myri10ge_check_statblock(struct myri10ge_priv *mgp)
{
	struct mcp_irq_data *stats = mgp->fw_stats;

	if (unlikely(stats->stats_updated)) {
		unsigned link_up = ntohl(stats->link_up);
		if (mgp->link_state != link_up) {
			mgp->link_state = link_up;
			
			if (mgp->link_state == MXGEFW_LINK_UP) {
				if (netif_msg_link(mgp))
					printk(KERN_INFO
					       "myri10ge: %s: link up\n",
					       mgp->dev->name);
				netif_carrier_on(mgp->dev);
				mgp->link_changes++;
			} else {
				if (netif_msg_link(mgp))
					printk(KERN_INFO
					       "myri10ge: %s: link %s\n",
					       mgp->dev->name,
					       (link_up == MXGEFW_LINK_MYRINET ? 
						"mismatch (Myrinet detected)" :
						"down"));
				netif_carrier_off(mgp->dev);
				mgp->link_changes++;
			}
		}
		if (mgp->rdma_tags_available !=
		    ntohl(mgp->fw_stats->rdma_tags_available)) {
			mgp->rdma_tags_available = ntohl(
				mgp->fw_stats->rdma_tags_available);
			printk(KERN_WARNING "myri10ge: %s: RDMA timed out! "
			       "%d tags left\n", mgp->dev->name,
			       mgp->rdma_tags_available);
		}
		mgp->down_cnt += stats->link_down;
		if (stats->link_down)
			wake_up(&mgp->down_wq);
	}
}

#ifdef MYRI10GE_NAPI
static int
myri10ge_poll(struct net_device *netdev, int *budget)
{
	struct myri10ge_priv *mgp = netdev_priv(netdev);
	struct myri10ge_rx_done *rx_done = &mgp->rx_done;
	int limit, orig_limit, work_done;

	/* process as many rx events as NAPI will allow */
	limit = min(*budget, netdev->quota);
	orig_limit = limit;
	myri10ge_clean_rx_done(mgp, &limit);
	work_done = orig_limit - limit;
	*budget -= work_done;
	netdev->quota -= work_done;

	if (rx_done->entry[rx_done->idx].length == 0 ||
	    !netif_running(netdev)) {
		netif_rx_complete(netdev);
		put_be32(htonl(3), mgp->irq_claim);
		return 0;
	}
	return 1;
}

static irqreturn_t
#ifdef MYRI10GE_HAVE_IRQ_HANDLER_REGS
myri10ge_intr(int irq __unused, void *arg, struct pt_regs *regs __unused)
#else /* ~MYRI10GE_HAVE_IRQ_HANDLER_REGS */
myri10ge_intr(int irq __unused, void *arg)
#endif /* ~MYRI10GE_HAVE_IRQ_HANDLER_REGS */
{
	struct myri10ge_priv *mgp = arg;
	struct mcp_irq_data *stats = mgp->fw_stats;
	struct myri10ge_tx_buf *tx = &mgp->tx;
	u32 send_done_count;
	int i;

	/* make sure it is our IRQ, and that the DMA has finished */
	if (unlikely(!stats->valid))
		return (IRQ_NONE);

	/* low bit indicates receives are present, so schedule
	   napi poll handler */
	if (stats->valid & 1)
		netif_rx_schedule(mgp->dev);

	if (!mgp->msi_enabled) {
		put_be32(0, mgp->irq_deassert);
		if (!myri10ge_deassert_wait)
			stats->valid = 0;
		mb();
	} else
		stats->valid = 0;


	/* Wait for IRQ line to go low, if using INTx */
	i = 0;
	while (1) {
		i++;
		/* check for transmit completes and receives */
		send_done_count = ntohl(stats->send_done_count);
		if (send_done_count != tx->pkt_done)
			myri10ge_tx_done(mgp, (int)send_done_count);
		if (unlikely(i > myri10ge_max_irq_loops)) {
			printk(KERN_WARNING "myri10ge: %s: irq stuck?\n",
			       mgp->dev->name);
			stats->valid = 0;
			schedule_work(&mgp->watchdog_work);
		}
		if (likely(stats->valid == 0))
			break;
		cpu_relax();
		barrier();
	}

	myri10ge_check_statblock(mgp);

	put_be32(htonl(3), mgp->irq_claim + 1);
	return (IRQ_HANDLED);
}
#else
static irqreturn_t
#ifdef MYRI10GE_HAVE_IRQ_HANDLER_REGS
myri10ge_intr(int irq __unused, void *arg, struct pt_regs *regs __unused)
#else /* ~MYRI10GE_HAVE_IRQ_HANDLER_REGS */
myri10ge_intr(int irq __unused, void *arg)
#endif /* ~MYRI10GE_HAVE_IRQ_HANDLER_REGS */
{
	struct myri10ge_priv *mgp = arg;
	struct mcp_irq_data *stats = mgp->fw_stats;
	struct myri10ge_tx_buf *tx = &mgp->tx;
	struct myri10ge_rx_done *rx_done = &mgp->rx_done;
	u32 send_done_count;
	int i, limit;
	u8 valid;

	/* make sure it is our IRQ, and that the DMA has finished */
	if (unlikely(!stats->valid))
		return (IRQ_NONE);

	valid = stats->valid;
	if (!mgp->msi_enabled) {
		put_be32(0, mgp->irq_deassert);
		if (!myri10ge_deassert_wait)
			stats->valid = 0;
		mb();
	} else
		stats->valid = 0;


	/* Wait for IRQ line to go low, if using INTx */
	i = 0;
	while (1) {
		i++;
		/* check for transmit completes and receives */
		send_done_count = ntohl(stats->send_done_count);
		while ((send_done_count != tx->pkt_done) ||
		       (rx_done->entry[rx_done->idx].length != 0)) {
			myri10ge_tx_done(mgp, (int)send_done_count);
			limit = 32;
			myri10ge_clean_rx_done(mgp, &limit);
		}
		if (unlikely(i > myri10ge_max_irq_loops)) {
			printk(KERN_WARNING "myri10ge: %s: irq stuck?\n",
			       mgp->dev->name);
			stats->valid = 0;
			schedule_work(&mgp->watchdog_work);
		}
		if (likely(stats->valid == 0))
			break;
		cpu_relax();
		barrier();
	}

	myri10ge_check_statblock(mgp);

	/* pass back rx token if we own it */
	if (valid & 0x1)
		put_be32(htonl(3), mgp->irq_claim);

	put_be32(htonl(3), mgp->irq_claim + 1);
	return (IRQ_HANDLED);
}

#endif /* MYRI10GE_NAPI */


static int
myri10ge_get_settings(struct net_device *netdev, struct ethtool_cmd *cmd)
{
	cmd->autoneg = AUTONEG_DISABLE;
	cmd->speed = SPEED_10000;
	cmd->duplex = DUPLEX_FULL;
	return 0;
}

static void
myri10ge_get_drvinfo(struct net_device *netdev,
		   struct ethtool_drvinfo *info)
{
	struct myri10ge_priv *mgp = netdev_priv(netdev);

	strlcpy(info->driver, "myri10ge", sizeof (info->driver));
	strlcpy(info->version, MYRI10GE_VERSION_STR, sizeof (info->version));
	strlcpy(info->fw_version, mgp->fw_version, sizeof (info->fw_version));
	strlcpy(info->bus_info, pci_name(mgp->pdev), sizeof (info->bus_info));
}

static int
myri10ge_get_coalesce(struct net_device *netdev,
		     struct ethtool_coalesce *coal)
{
	struct myri10ge_priv *mgp = netdev_priv(netdev);
	coal->rx_coalesce_usecs = mgp->intr_coal_delay;
	return 0;
}

static int
myri10ge_set_coalesce(struct net_device *netdev,
		     struct ethtool_coalesce *coal)
{
	struct myri10ge_priv *mgp = netdev_priv(netdev);

	mgp->intr_coal_delay = coal->rx_coalesce_usecs;
	put_be32(htonl(mgp->intr_coal_delay), mgp->intr_coal_delay_ptr);
	return 0;
}

static void
myri10ge_get_pauseparam(struct net_device *netdev,
			struct ethtool_pauseparam *pause)
{
	struct myri10ge_priv *mgp = netdev_priv(netdev);

	pause->autoneg = 0;
	pause->rx_pause = mgp->pause;
	pause->tx_pause = mgp->pause;
}

static int
myri10ge_set_pauseparam(struct net_device *netdev,
			struct ethtool_pauseparam *pause)
{
	struct myri10ge_priv *mgp = netdev_priv(netdev);

	if (pause->tx_pause != mgp->pause)
		return myri10ge_change_pause(mgp, pause->tx_pause);
	if (pause->rx_pause != mgp->pause)
		return myri10ge_change_pause(mgp, pause->tx_pause);
	if (pause->autoneg != 0)
		return -EINVAL;
	return 0;
}

static void
myri10ge_get_ringparam(struct net_device *netdev,
		       struct ethtool_ringparam *ring)
{
	struct myri10ge_priv *mgp = netdev_priv(netdev);

	ring->rx_mini_max_pending = mgp->rx_small.mask + 1;
	ring->rx_max_pending = mgp->rx_big.mask + 1;
	ring->rx_jumbo_max_pending = 0;
	ring->tx_max_pending = mgp->rx_small.mask + 1;
	ring->rx_mini_pending = ring->rx_mini_max_pending;
	ring->rx_pending = ring->rx_max_pending;
	ring->rx_jumbo_pending = ring->rx_jumbo_max_pending;
	ring->tx_pending = ring->tx_max_pending;
}

static u32
myri10ge_get_rx_csum(struct net_device *netdev)
{
	struct myri10ge_priv *mgp = netdev_priv(netdev);
	if (mgp->csum_flag)
		return 1;
	else
		return 0;
}

static int
myri10ge_set_rx_csum(struct net_device *netdev, u32 csum_enabled)
{
	struct myri10ge_priv *mgp = netdev_priv(netdev);
	if (csum_enabled)
		mgp->csum_flag = MXGEFW_FLAGS_CKSUM;
	else
		mgp->csum_flag = 0;
	return 0;
}

#ifndef LINUX_KERNEL_SPECIFIC

static u32
myri10ge_get_tx_csum(struct net_device *netdev)
{
	if ((netdev->features & NETIF_F_HW_CSUM) != 0)
		return 1;
	else
		return 0;
}

static int
myri10ge_set_tx_csum(struct net_device *netdev, u32 csum_enabled)
{
	if (csum_enabled)
		netdev->features |= NETIF_F_HW_CSUM;
	else
		netdev->features &= ~NETIF_F_HW_CSUM;
	return 0;
}
#endif


static const char myri10ge_gstrings_stats[][ETH_GSTRING_LEN] = {
	"rx_packets", "tx_packets", "rx_bytes", "tx_bytes", "rx_errors",
	"tx_errors", "rx_dropped", "tx_dropped", "multicast", "collisions",
	"rx_length_errors", "rx_over_errors", "rx_crc_errors",
	"rx_frame_errors", "rx_fifo_errors", "rx_missed_errors",
	"tx_aborted_errors", "tx_carrier_errors", "tx_fifo_errors",
	"tx_heartbeat_errors", "tx_window_errors",
	/* device-specific stats */
#ifndef LINUX_KERNEL_SPECIFIC
	"rx_skbs", "alloc_order", "builtin_fw", "napi",
#endif
	"tx_boundary", "WC", "irq", "MSI",
	"read_dma_bw_MBs", "write_dma_bw_MBs", "read_write_dma_bw_MBs",
	"serial_number", "tx_pkt_start", "tx_pkt_done",
	"tx_req", "tx_done", "rx_small_cnt", "rx_big_cnt",
#if MYRI10GE_LRO
	"lro_queued", "lro_flushed", "lro_too_many_streams", "lro_bad_csum",
#endif
	"wake_queue", "stop_queue", "watchdog_resets", "tx_linearized",
	"link_changes", "link_up", "dropped_link_overflow", 
	"dropped_link_error_or_filtered",
	"dropped_pause", "dropped_bad_phy", "dropped_bad_crc32",	
	"dropped_unicast_filtered", "dropped_multicast_filtered",
	"dropped_runt", "dropped_overrun", "dropped_no_small_buffer",
	"dropped_no_big_buffer"
};
#define MYRI10GE_NET_STATS_LEN      21
#define MYRI10GE_STATS_LEN  sizeof(myri10ge_gstrings_stats) / ETH_GSTRING_LEN

static void
myri10ge_get_strings(struct net_device *netdev, u32 stringset, u8 *data)
{
	switch (stringset) {
	case ETH_SS_STATS:
		memcpy(data, *myri10ge_gstrings_stats,
		       sizeof(myri10ge_gstrings_stats));
		break;
	}
}

static int
myri10ge_get_stats_count(struct net_device *netdev)
{
	return MYRI10GE_STATS_LEN;
}

static void
myri10ge_get_ethtool_stats(struct net_device *netdev,
			   struct ethtool_stats *stats, u64 *data)
{
	struct myri10ge_priv *mgp = netdev_priv(netdev);
	int i;

	for(i = 0; i < MYRI10GE_NET_STATS_LEN; i++)
		data[i] = ((unsigned long *) &mgp->stats)[i];

#ifndef LINUX_KERNEL_SPECIFIC
	data[i++] = MYRI10GE_RX_SKBS;
	data[i++] = MYRI10GE_ALLOC_ORDER;
	data[i++] = MYRI10GE_BUILTIN_FW;
#ifdef MYRI10GE_NAPI
	data[i++] = 1;
#else
	data[i++] = 0;
#endif
#endif
	data[i++] = (unsigned int)mgp->tx.boundary;
	data[i++] = (unsigned int)mgp->wc_enabled;
	data[i++] = (unsigned int)mgp->pdev->irq;
	data[i++] = (unsigned int)mgp->msi_enabled;
	data[i++] = (unsigned int)mgp->read_dma;
	data[i++] = (unsigned int)mgp->write_dma;
	data[i++] = (unsigned int)mgp->read_write_dma;
	data[i++] = (unsigned int)mgp->serial_number;
	data[i++] = (unsigned int)mgp->tx.pkt_start;
	data[i++] = (unsigned int)mgp->tx.pkt_done;
	data[i++] = (unsigned int)mgp->tx.req;
	data[i++] = (unsigned int)mgp->tx.done;
	data[i++] = (unsigned int)mgp->rx_small.cnt;
	data[i++] = (unsigned int)mgp->rx_big.cnt;
#if MYRI10GE_LRO
	data[i++] = (unsigned int)mgp->lro_queued;
	data[i++] = (unsigned int)mgp->lro_flushed;
	data[i++] = (unsigned int)mgp->lro_too_many_streams;
	data[i++] = (unsigned int)mgp->lro_bad_csum;
#endif
	data[i++] = (unsigned int)mgp->wake_queue;
	data[i++] = (unsigned int)mgp->stop_queue;
	data[i++] = (unsigned int)mgp->watchdog_resets;
	data[i++] = (unsigned int)mgp->tx_linearized;
	data[i++] = (unsigned int)mgp->link_changes;
	data[i++] = (unsigned int)ntohl(mgp->fw_stats->link_up);
	data[i++] = (unsigned int)ntohl(mgp->fw_stats->dropped_link_overflow);
	data[i++] = (unsigned int)ntohl(mgp->fw_stats->dropped_link_error_or_filtered);
	data[i++] = (unsigned int)ntohl(mgp->fw_stats->dropped_pause);
	data[i++] = (unsigned int)ntohl(mgp->fw_stats->dropped_bad_phy);
	data[i++] = (unsigned int)ntohl(mgp->fw_stats->dropped_bad_crc32);
	data[i++] = (unsigned int)ntohl(mgp->fw_stats->dropped_unicast_filtered);
	data[i++] = (unsigned int)ntohl(mgp->fw_stats->dropped_multicast_filtered);
	data[i++] = (unsigned int)ntohl(mgp->fw_stats->dropped_runt);
	data[i++] = (unsigned int)ntohl(mgp->fw_stats->dropped_overrun);
	data[i++] = (unsigned int)ntohl(mgp->fw_stats->dropped_no_small_buffer);
	data[i++] = (unsigned int)ntohl(mgp->fw_stats->dropped_no_big_buffer);
}

static void myri10ge_set_msglevel(struct net_device *netdev, u32 value)
{
	struct myri10ge_priv *mgp = netdev_priv(netdev);
	mgp->msg_enable = value;
}

static u32 myri10ge_get_msglevel(struct net_device *netdev)
{
	struct myri10ge_priv *mgp = netdev_priv(netdev);
	return mgp->msg_enable;
}

static MYRI10GE_ETHTOOL_OPS_TYPE myri10ge_ethtool_ops = {
	.get_settings 			= myri10ge_get_settings,
	.get_drvinfo			= myri10ge_get_drvinfo,
	.get_coalesce			= myri10ge_get_coalesce,
	.set_coalesce			= myri10ge_set_coalesce,
	.get_pauseparam			= myri10ge_get_pauseparam,
	.set_pauseparam			= myri10ge_set_pauseparam,
	.get_ringparam			= myri10ge_get_ringparam,
	.get_rx_csum			= myri10ge_get_rx_csum,
	.set_rx_csum			= myri10ge_set_rx_csum,
#ifdef LINUX_KERNEL_SPECIFIC
	.get_tx_csum			= ethtool_op_get_tx_csum,
	.set_tx_csum			= ethtool_op_set_tx_hw_csum,
#else
	.get_tx_csum			= myri10ge_get_tx_csum,
	.set_tx_csum			= myri10ge_set_tx_csum,
#endif
	.get_sg				= ethtool_op_get_sg,
	.set_sg				= ethtool_op_set_sg,
	.get_tso			= ethtool_op_get_tso,
	.set_tso			= ethtool_op_set_tso,
	.get_link			= ethtool_op_get_link,
	.get_strings			= myri10ge_get_strings,
	.get_stats_count		= myri10ge_get_stats_count,
	.get_ethtool_stats		= myri10ge_get_ethtool_stats,
	.set_msglevel			= myri10ge_set_msglevel,
	.get_msglevel			= myri10ge_get_msglevel
};

static int
myri10ge_allocate_rings(struct net_device *dev)
{
	struct myri10ge_priv *mgp;
	struct myri10ge_cmd cmd;
	int tx_ring_size, rx_ring_size;
	int tx_ring_entries, rx_ring_entries;
	int i, status;
	size_t bytes;

	mgp = netdev_priv(dev);

	/* get ring sizes */

	status = myri10ge_send_cmd(mgp, MXGEFW_CMD_GET_SEND_RING_SIZE, &cmd, 0);
	tx_ring_size = cmd.data0;
	status |= myri10ge_send_cmd(mgp, MXGEFW_CMD_GET_RX_RING_SIZE, &cmd, 0);
	if (status != 0)
		return status;
	rx_ring_size = cmd.data0;

	tx_ring_entries = tx_ring_size / sizeof (struct mcp_kreq_ether_send);
	rx_ring_entries = rx_ring_size / sizeof (struct mcp_dma_addr);
	mgp->tx.mask = tx_ring_entries - 1;
	mgp->rx_small.mask = mgp->rx_big.mask = rx_ring_entries - 1;

	status = -ENOMEM;

	/* allocate the host shadow rings */

	bytes = 8 + (MYRI10GE_MAX_SEND_DESC_TSO + 4)
		* sizeof (*mgp->tx.req_list);
	mgp->tx.req_bytes = myri10ge_kzalloc(bytes, GFP_KERNEL);
	if (mgp->tx.req_bytes == NULL)
		goto abort_with_nothing;

	/* ensure req_list entries are aligned to 8 bytes */
	mgp->tx.req_list = (struct mcp_kreq_ether_send *)
		ALIGN((unsigned long) mgp->tx.req_bytes, 8);

	bytes = rx_ring_entries * sizeof (*mgp->rx_small.shadow);
	mgp->rx_small.shadow = myri10ge_kzalloc(bytes, GFP_KERNEL);
	if (mgp->rx_small.shadow == NULL)
		goto abort_with_tx_req_bytes;

	bytes = rx_ring_entries * sizeof (*mgp->rx_big.shadow);
	mgp->rx_big.shadow = myri10ge_kzalloc(bytes, GFP_KERNEL);
	if (mgp->rx_big.shadow == NULL)
		goto abort_with_rx_small_shadow;

	/* allocate the host info rings */

	bytes = tx_ring_entries * sizeof (*mgp->tx.info);
	mgp->tx.info = myri10ge_kzalloc(bytes, GFP_KERNEL);
	if (mgp->tx.info == NULL)
		goto abort_with_rx_big_shadow;

	bytes = rx_ring_entries * sizeof (*mgp->rx_small.info);
	mgp->rx_small.info = myri10ge_kzalloc(bytes, GFP_KERNEL);
	if (mgp->rx_small.info == NULL)
		goto abort_with_tx_info;

	bytes = rx_ring_entries * sizeof (*mgp->rx_big.info);
	mgp->rx_big.info = myri10ge_kzalloc(bytes, GFP_KERNEL);
	if (mgp->rx_big.info == NULL)
		goto abort_with_rx_small_info;

	/* Fill the receive rings */
	mgp->rx_big.cnt = 0;
	mgp->rx_small.cnt = 0;
	mgp->rx_big.fill_cnt = 0;
	mgp->rx_small.fill_cnt = 0;
	mgp->rx_small.page_offset = MYRI10GE_ALLOC_SIZE;
	mgp->rx_big.page_offset = MYRI10GE_ALLOC_SIZE;
	mgp->rx_small.watchdog_needed = 0;
	mgp->rx_big.watchdog_needed = 0;
#if MYRI10GE_RX_SKBS
	for (i = 0; i <= mgp->rx_small.mask; i++) {
		status = myri10ge_getbuf(&mgp->rx_small, mgp,
					 mgp->small_bytes, i);
		if (status) {
			printk(KERN_ERR "myri10ge: %s: alloced only %d small bufs\n",
			       dev->name, i);
			goto abort_with_rx_small_ring;
		}
	}

	for (i = 0; i <= mgp->rx_big.mask; i++) {
		status = myri10ge_getbuf(&mgp->rx_big, mgp, mgp->big_bytes, i);
		if (status) {
			printk(KERN_ERR "myri10ge: %s: alloced only %d big bufs\n",
			       dev->name, i);
			goto abort_with_rx_big_ring;
		}
	}
#else /* MYRI10GE_RX_SKBS */
	myri10ge_alloc_rx_pages(mgp, &mgp->rx_small, 
				mgp->small_bytes + MXGEFW_PAD, 0);

	if (mgp->rx_small.fill_cnt < mgp->rx_small.mask + 1) {
		printk(KERN_ERR "myri10ge: %s: alloced only %d small bufs\n",
		       dev->name, mgp->rx_small.fill_cnt);
		goto abort_with_rx_small_ring;
	}

	myri10ge_alloc_rx_pages(mgp, &mgp->rx_big, mgp->big_bytes, 0);
	if (mgp->rx_big.fill_cnt < mgp->rx_big.mask + 1) {
		printk(KERN_ERR "myri10ge: %s: alloced only %d big bufs\n",
		       dev->name, mgp->rx_big.fill_cnt);
		goto abort_with_rx_big_ring;
	}
#endif

#if MYRI10GE_LRO
	bytes = sizeof (struct myri10ge_lro_packet);
	INIT_HLIST_HEAD(&mgp->rx_done.lro_free);
	INIT_HLIST_HEAD(&mgp->rx_done.lro_active);
	for (i = 0; i < myri10ge_lro; i++) {
		struct myri10ge_lro_packet *lro;
		lro = myri10ge_kzalloc(bytes, GFP_KERNEL);
		if (lro != NULL) {
			INIT_HLIST_NODE(&lro->lro_node);
			hlist_add_head(&lro->lro_node, &mgp->rx_done.lro_free);
		}
	}
#endif /* MYRI10GE_LRO */

	return 0;


#if MYRI10GE_RX_SKBS
abort_with_rx_big_ring:
	for (i = 0; i <= mgp->rx_big.mask; i++) {
		if (mgp->rx_big.info[i].skb != NULL)
			dev_kfree_skb_any(mgp->rx_big.info[i].skb);
		if (pci_unmap_len(&mgp->rx_big.info[i], len))
			pci_unmap_single(mgp->pdev,
					 pci_unmap_addr(&mgp->rx_big.info[i], bus),
					 pci_unmap_len(&mgp->rx_big.info[i], len),
					 PCI_DMA_FROMDEVICE);
	}

abort_with_rx_small_ring:
	for (i = 0; i <= mgp->rx_small.mask; i++) {
		if (mgp->rx_small.info[i].skb != NULL)
			dev_kfree_skb_any(mgp->rx_small.info[i].skb);
		if (pci_unmap_len(&mgp->rx_small.info[i], len))
			pci_unmap_single(mgp->pdev,
					 pci_unmap_addr(&mgp->rx_small.info[i], bus),
					 pci_unmap_len(&mgp->rx_small.info[i], len),
					 PCI_DMA_FROMDEVICE);
	}
#else
abort_with_rx_big_ring:
	for (i = mgp->rx_big.cnt; i< mgp->rx_big.fill_cnt; i++) {
		int idx = i & mgp->rx_big.mask;
		myri10ge_unmap_rx_page(mgp->pdev, &mgp->rx_big.info[idx],
				       mgp->big_bytes);
		put_page(mgp->rx_big.info[idx].page);
	}

abort_with_rx_small_ring:
	for (i = mgp->rx_small.cnt; i< mgp->rx_small.fill_cnt; i++) {
		int idx = i & mgp->rx_small.mask;
		myri10ge_unmap_rx_page(mgp->pdev, &mgp->rx_small.info[idx],
				       mgp->small_bytes + MXGEFW_PAD);
		put_page(mgp->rx_small.info[idx].page);
	}
#endif /* MYRI10GE_RX_SKBS */

	kfree(mgp->rx_big.info);

abort_with_rx_small_info:
	kfree(mgp->rx_small.info);

abort_with_tx_info:
	kfree(mgp->tx.info);

abort_with_rx_big_shadow:
	kfree(mgp->rx_big.shadow);

abort_with_rx_small_shadow:
	kfree(mgp->rx_small.shadow);

abort_with_tx_req_bytes:
	kfree(mgp->tx.req_bytes);
	mgp->tx.req_bytes = NULL;
	mgp->tx.req_list = NULL;

abort_with_nothing:
	return status;
}

static void
myri10ge_free_rings(struct net_device *dev)
{
	struct myri10ge_priv *mgp;
	struct sk_buff *skb;
	struct myri10ge_tx_buf *tx;
#if MYRI10GE_LRO
	struct hlist_node *node, *node2;
	struct myri10ge_lro_packet *lro;
#endif
	int i, len, idx;

	mgp = netdev_priv(dev);

#if MYRI10GE_LRO
	hlist_for_each_entry_safe(lro, node, node2, &mgp->rx_done.lro_free, lro_node) {
		hlist_del(&lro->lro_node);
		kfree(lro);
	}
#endif /* MYRI10GE_LRO */



#if MYRI10GE_RX_SKBS
	for (i = 0; i <= mgp->rx_big.mask; i++) {
		if (mgp->rx_big.info[i].skb != NULL)
			dev_kfree_skb_any(mgp->rx_big.info[i].skb);
		if (pci_unmap_len(&mgp->rx_big.info[i], len))
			pci_unmap_single(mgp->pdev,
					 pci_unmap_addr(&mgp->rx_big.info[i], bus),
					 pci_unmap_len(&mgp->rx_big.info[i], len),
					 PCI_DMA_FROMDEVICE);
	}

	for (i = 0; i <= mgp->rx_small.mask; i++) {
		if (mgp->rx_small.info[i].skb != NULL)
			dev_kfree_skb_any(mgp->rx_small.info[i].skb);
		if (pci_unmap_len(&mgp->rx_small.info[i], len))
			pci_unmap_single(mgp->pdev,
					 pci_unmap_addr(&mgp->rx_small.info[i], bus),
					 pci_unmap_len(&mgp->rx_small.info[i], len),
					 PCI_DMA_FROMDEVICE);
	}
#else	
	for (i = mgp->rx_big.cnt; i< mgp->rx_big.fill_cnt; i++) {
		idx = i & mgp->rx_big.mask;
		if (i == mgp->rx_big.fill_cnt - 1)
			mgp->rx_big.info[idx].page_offset = MYRI10GE_ALLOC_SIZE;
		myri10ge_unmap_rx_page(mgp->pdev, &mgp->rx_big.info[idx],
				       mgp->big_bytes);
		put_page(mgp->rx_big.info[idx].page);
	}

	for (i = mgp->rx_small.cnt; i< mgp->rx_small.fill_cnt; i++) {
		idx = i & mgp->rx_small.mask;
		if (i == mgp->rx_small.fill_cnt - 1)
			mgp->rx_small.info[idx].page_offset = MYRI10GE_ALLOC_SIZE;
		myri10ge_unmap_rx_page(mgp->pdev, &mgp->rx_small.info[idx],
				       mgp->small_bytes + MXGEFW_PAD);
		put_page(mgp->rx_small.info[idx].page);
	}
#endif /* MYRI10GE_RX_SKBS */
	tx = &mgp->tx;
	while (tx->done != tx->req) {
		idx = tx->done & tx->mask;
		skb = tx->info[idx].skb;

		/* Mark as free */
		tx->info[idx].skb = NULL;
		tx->done++;
		len = pci_unmap_len(&tx->info[idx], len);
		pci_unmap_len_set(&tx->info[idx], len, 0);
		if (skb) {
			mgp->stats.tx_dropped++;
			dev_kfree_skb_any(skb);
			if (len)
				pci_unmap_single(mgp->pdev,
						 pci_unmap_addr(&tx->info[idx], bus),
						 len, PCI_DMA_TODEVICE);
		} else {
			if (len)
				pci_unmap_page(mgp->pdev,
					       pci_unmap_addr(&tx->info[idx], bus),
					       len, PCI_DMA_TODEVICE);
		}
	}
	kfree(mgp->rx_big.info);

	kfree(mgp->rx_small.info);

	kfree(mgp->tx.info);

	kfree(mgp->rx_big.shadow);

	kfree(mgp->rx_small.shadow);

	kfree(mgp->tx.req_bytes);
	mgp->tx.req_bytes = NULL;
	mgp->tx.req_list = NULL;
}

static int
myri10ge_request_irq(struct myri10ge_priv *mgp)
{
	struct pci_dev *pdev = mgp->pdev;
	int status;

#ifdef MYRI10GE_HAVE_MSI
	if (myri10ge_use_msi(pdev)) {
		status = pci_enable_msi(pdev);
		if (status != 0)
			dev_err(&pdev->dev,
				"Error %d setting up MSI; falling back to xPIC\n",
				status);
		else
			mgp->msi_enabled = 1;
	} else {
		mgp->msi_enabled = 0;
	}
#endif
	status = request_irq(pdev->irq, myri10ge_intr, IRQF_SHARED,
			     mgp->dev->name, mgp);
	if (status != 0) {
		dev_err(&pdev->dev, "failed to allocate IRQ\n");
#ifdef MYRI10GE_HAVE_MSI
		if (mgp->msi_enabled)
			pci_disable_msi(pdev);
#endif
	}
	return status;
}

static void
myri10ge_free_irq(struct myri10ge_priv *mgp)
{
	struct pci_dev *pdev = mgp->pdev;

	free_irq(pdev->irq, mgp);
#ifdef MYRI10GE_HAVE_MSI
	if (mgp->msi_enabled)
		pci_disable_msi(pdev);
#endif
}

static int
myri10ge_open(struct net_device *dev)
{
	struct myri10ge_priv *mgp;
	struct myri10ge_cmd cmd;
	int status, big_pow2;

	mgp = netdev_priv(dev);

	if (mgp->running != MYRI10GE_ETH_STOPPED)
		return -EBUSY;

	mgp->running = MYRI10GE_ETH_STARTING;
	status = myri10ge_reset(mgp);
	if (status != 0) {
		printk(KERN_ERR "myri10ge: %s: failed reset\n", dev->name);
		goto abort_with_nothing;
	}

	status = myri10ge_request_irq(mgp);
	if (status != 0)
		goto abort_with_nothing;

	/* decide what small buffer size to use.  For good TCP rx
	 * performance, it is important to not receive 1514 byte
	 * frames into jumbo buffers, as it confuses the socket buffer
	 * accounting code, leading to drops and erratic performance.
	 */

	if (dev->mtu <= ETH_DATA_LEN)
		/* enough for a TCP header */
		mgp->small_bytes = (128 > SMP_CACHE_BYTES) 
			? (128 - MXGEFW_PAD)
			: (SMP_CACHE_BYTES - MXGEFW_PAD); 
	else
		/* enough for a vlan encapsulated ETH_DATA_LEN frame */
		mgp->small_bytes = VLAN_ETH_FRAME_LEN;

	/* Override the small buffer size? */
	if (myri10ge_small_bytes > 0)
		mgp->small_bytes = myri10ge_small_bytes;


	/* get the lanai pointers to the send and receive rings */

	status |= myri10ge_send_cmd(mgp, MXGEFW_CMD_GET_SEND_OFFSET, &cmd, 0);
	mgp->tx.lanai = (struct mcp_kreq_ether_send __iomem *) (mgp->sram + cmd.data0);

	status |= myri10ge_send_cmd(mgp, MXGEFW_CMD_GET_SMALL_RX_OFFSET, &cmd, 0);
	mgp->rx_small.lanai = (struct mcp_kreq_ether_recv __iomem *) (mgp->sram + cmd.data0);

	status |= myri10ge_send_cmd(mgp, MXGEFW_CMD_GET_BIG_RX_OFFSET, &cmd, 0);
	mgp->rx_big.lanai = (struct mcp_kreq_ether_recv __iomem *) (mgp->sram + cmd.data0);

	if (status != 0) {
		printk(KERN_ERR "myri10ge: %s: failed to get ring sizes or locations\n",
		      dev->name);
		mgp->running = MYRI10GE_ETH_STOPPED;
		goto abort_with_irq;
	}

	if (myri10ge_wcfifo && mgp->wc_enabled) {
		mgp->tx.wc_fifo = (u8 __iomem *) mgp->sram + MXGEFW_ETH_SEND_4;
		mgp->rx_small.wc_fifo = (u8 __iomem *) mgp->sram + MXGEFW_ETH_RECV_SMALL;
		mgp->rx_big.wc_fifo = (u8 __iomem *) mgp->sram + MXGEFW_ETH_RECV_BIG;
	} else {
		mgp->tx.wc_fifo = NULL;
		mgp->rx_small.wc_fifo = NULL;
		mgp->rx_big.wc_fifo = NULL;
	}

	/* Firmware needs the big buff size as a power of 2.  Lie and
	 * tell him the buffer is larger, because we only use 1
	 * buffer/pkt, and the mtu will prevent overruns.
	 */
	big_pow2 = dev->mtu + ETH_HLEN + VLAN_HLEN + MXGEFW_PAD;
#if MYRI10GE_RX_SKBS
	while (!myri10ge_is_power_of_2(big_pow2))
		big_pow2++;
	mgp->big_bytes = dev->mtu + ETH_HLEN + VLAN_HLEN + MXGEFW_PAD;
#else
	if (big_pow2 < MYRI10GE_ALLOC_SIZE/2) {
		while (!myri10ge_is_power_of_2(big_pow2))
			big_pow2++;
		mgp->big_bytes = dev->mtu + ETH_HLEN + VLAN_HLEN + MXGEFW_PAD;
	} else {
		big_pow2 = MYRI10GE_ALLOC_SIZE;
		mgp->big_bytes = big_pow2;
	}
#endif

	status = myri10ge_allocate_rings(dev);
	if (status != 0)
		goto abort_with_irq;

	/* now give firmware buffers sizes, and MTU */
	cmd.data0 = dev->mtu + ETH_HLEN + VLAN_HLEN;
	status = myri10ge_send_cmd(mgp, MXGEFW_CMD_SET_MTU, &cmd, 0);
	cmd.data0 = mgp->small_bytes;
	status |= myri10ge_send_cmd(mgp, MXGEFW_CMD_SET_SMALL_BUFFER_SIZE, &cmd, 0);
	cmd.data0 = big_pow2;
	status |= myri10ge_send_cmd(mgp, MXGEFW_CMD_SET_BIG_BUFFER_SIZE, &cmd, 0);
	if (status) {
		printk(KERN_ERR "myri10ge: %s: Couldn't set buffer sizes\n",
		       dev->name);
		goto abort_with_rings;
	}

	cmd.data0 = MYRI10GE_LOWPART_TO_U32(mgp->fw_stats_bus);
	cmd.data1 = MYRI10GE_HIGHPART_TO_U32(mgp->fw_stats_bus);
	cmd.data2 = sizeof(struct mcp_irq_data);
	status = myri10ge_send_cmd(mgp, MXGEFW_CMD_SET_STATS_DMA_V2, &cmd, 0);
	if (status == -ENOSYS) {
		dma_addr_t bus = mgp->fw_stats_bus;
		bus += offsetof(struct mcp_irq_data, send_done_count);
		cmd.data0 = MYRI10GE_LOWPART_TO_U32(bus);
		cmd.data1 = MYRI10GE_HIGHPART_TO_U32(bus);
		status = myri10ge_send_cmd(mgp, 
					   MXGEFW_CMD_SET_STATS_DMA_OBSOLETE,
					   &cmd, 0);
		/* Firmware cannot support multicast without STATS_DMA_V2 */
		mgp->fw_multicast_support = 0;
	} else {
		mgp->fw_multicast_support = 1;
	}
	if (status) {
		printk(KERN_ERR "myri10ge: %s: Couldn't set stats DMA\n",
		       dev->name);
		goto abort_with_rings;
	}

	mgp->link_state = htonl(~0U);
	mgp->rdma_tags_available = 15;

#ifdef MYRI10GE_NAPI
	netif_poll_enable(mgp->dev); /* must happen prior to any irq */
#endif

	status = myri10ge_send_cmd(mgp, MXGEFW_CMD_ETHERNET_UP, &cmd, 0);
	if (status) {
		printk(KERN_ERR "myri10ge: %s: Couldn't bring up link\n",
		       dev->name);
		goto abort_with_rings;
	}

	mgp->wake_queue = 0;
	mgp->stop_queue = 0;
	mgp->running = MYRI10GE_ETH_RUNNING;
	mgp->watchdog_timer.expires =
		jiffies + myri10ge_watchdog_timeout * HZ;
	add_timer(&mgp->watchdog_timer);
	netif_wake_queue(dev);
	return 0;

abort_with_rings:
	myri10ge_free_rings(dev);

abort_with_irq:
	myri10ge_free_irq(mgp);

abort_with_nothing:
	mgp->running = MYRI10GE_ETH_STOPPED;
	return -ENOMEM;
}

static int
myri10ge_close(struct net_device *dev)
{
	struct myri10ge_priv *mgp;
	struct myri10ge_cmd cmd;
	int status, old_down_cnt;

	mgp = netdev_priv(dev);

	if (mgp->running != MYRI10GE_ETH_RUNNING)
		return 0;

	if (mgp->tx.req_bytes == NULL)
		return 0;

	del_timer_sync(&mgp->watchdog_timer);
	mgp->running = MYRI10GE_ETH_STOPPING;
#ifdef MYRI10GE_NAPI
	netif_poll_disable(mgp->dev);
#endif
	netif_carrier_off(dev);
	netif_stop_queue(dev);
	old_down_cnt = mgp->down_cnt;
	mb();
	status = myri10ge_send_cmd(mgp, MXGEFW_CMD_ETHERNET_DOWN, &cmd, 0);
	if (status)
		printk(KERN_ERR "myri10ge: %s: Couldn't bring down link\n",
		       dev->name);

#ifdef LINUX_KERNEL_SPECIFIC
	wait_event_timeout(mgp->down_wq, old_down_cnt != mgp->down_cnt, HZ);
#else
	if (old_down_cnt == mgp->down_cnt)
		myri10ge_msleep(1000);
#endif
	if (old_down_cnt == mgp->down_cnt)
		printk(KERN_ERR "myri10ge: %s never got down irq\n",
		       dev->name);

	netif_tx_disable(dev);
	myri10ge_free_irq(mgp);
	myri10ge_free_rings(dev);

	mgp->running = MYRI10GE_ETH_STOPPED;
	return 0;
}

/* copy an array of struct mcp_kreq_ether_send's to the mcp.  Copy
 * backwards one at a time and handle ring wraps */

static inline void
myri10ge_submit_req_backwards(struct myri10ge_tx_buf *tx,
			      struct mcp_kreq_ether_send *src, int cnt)
{
	int idx, starting_slot;
	starting_slot = tx->req;
	while (cnt > 1) {
		cnt--;
		idx = (starting_slot + cnt) & tx->mask;
		myri10ge_pio_copy(&tx->lanai[idx],
				  &src[cnt], sizeof(*src));
		mb();
	}
}

/*
 * copy an array of struct mcp_kreq_ether_send's to the mcp.  Copy
 * at most 32 bytes at a time, so as to avoid involving the software
 * pio handler in the nic.   We re-write the first segment's flags
 * to mark them valid only after writing the entire chain.
 */

static inline void
myri10ge_submit_req(struct myri10ge_tx_buf *tx, struct mcp_kreq_ether_send *src,
		    int cnt)
{
	int idx, i;
	struct mcp_kreq_ether_send __iomem *dstp, *dst;
	struct mcp_kreq_ether_send *srcp;
	u8 last_flags;

	idx = tx->req & tx->mask;

	last_flags = src->flags;
	src->flags = 0;
	mb();
	dst = dstp = &tx->lanai[idx];
	srcp = src;

	if ((idx + cnt) < tx->mask) {
		for (i = 0; i < (cnt - 1); i += 2) {
			myri10ge_pio_copy(dstp, srcp, 2 * sizeof(*src));
			mb();	/* force write every 32 bytes */
			srcp += 2;
			dstp += 2;
		}
	} else {
		/* submit all but the first request, and ensure
		   that it is submitted below */
		myri10ge_submit_req_backwards(tx, src, cnt);
		i = 0;
	}
	if (i < cnt) {
		/* submit the first request */
		myri10ge_pio_copy(dstp, srcp, sizeof(*src));
		mb(); /* barrier before setting valid flag */
	}

	/* re-write the last 32-bits with the valid flags */
	src->flags = last_flags;
	put_be32(*((__be32 *) src + 3), (__be32 __iomem *) dst + 3);
	tx->req += cnt;
	mb();
}

static inline void
myri10ge_submit_req_wc(struct myri10ge_tx_buf *tx,
		       struct mcp_kreq_ether_send *src, int cnt)
{
	tx->req += cnt;
	mb();
	while (cnt >= 4) {
		myri10ge_pio_copy(tx->wc_fifo, src, 64);
		mb();
		src += 4;
		cnt -= 4;
	}
	if (cnt > 0) {
		/* pad it to 64 bytes.  The src is 64 bytes bigger than it
		 * needs to be so that we don't overrun it */
		myri10ge_pio_copy(tx->wc_fifo + MXGEFW_ETH_SEND_OFFSET(cnt), src, 64);
		mb();
	}
}

#ifndef LINUX_KERNEL_SPECIFIC
static void
myri10ge_csum_fixup(struct sk_buff *skb, int cksum_offset,
		    int pseudo_hdr_offset)
{
	int csum;
	u16 *csum_ptr;


	csum = skb_checksum(skb, cksum_offset,
			    skb->len - cksum_offset, 0);
	csum_ptr = (u16 *) (myri10ge_skb_transport_header(skb) + skb->csum);
	if (!pskb_may_pull(skb, pseudo_hdr_offset)) {
		printk(KERN_ERR "myri10ge: can't pull skb %d\n",
		       pseudo_hdr_offset);
		return;
	}
	*csum_ptr = csum_fold(csum);
	/* need to fixup IPv4 UDP packets according to RFC768 */
	if (unlikely(*csum_ptr == 0 &&
		     skb->protocol == htons(ETH_P_IP) &&
		     myri10ge_ip_hdr(skb)->protocol == IPPROTO_UDP))
		*csum_ptr = 0xffff;
}
#endif

/*
 * Transmit a packet.  We need to split the packet so that a single
 * segment does not cross myri10ge->tx.boundary, so this makes segment
 * counting tricky.  So rather than try to count segments up front, we
 * just give up if there are too few segments to hold a reasonably
 * fragmented packet currently available.  If we run
 * out of segments while preparing a packet for DMA, we just linearize
 * it and try again.
 */

static int
myri10ge_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct myri10ge_priv *mgp = netdev_priv(dev);
	struct mcp_kreq_ether_send *req;
	struct myri10ge_tx_buf *tx = &mgp->tx;
	struct skb_frag_struct *frag;
	dma_addr_t bus;
	u32 low;
	__be32 high_swapped;
	unsigned int len;
	int idx, last_idx, avail, frag_cnt, frag_idx, count, mss, max_segments;
	u16 pseudo_hdr_offset, cksum_offset;
	int cum_len, seglen, boundary, rdma_count;
	u8 flags, odd_flag;

again:
	req = tx->req_list;
	avail = tx->mask - 1 - (tx->req - tx->done);

	mss = 0;
	max_segments = MXGEFW_MAX_SEND_DESC;

	if (myri10ge_skb_is_gso(skb)) {
		mss = skb_shinfo(skb)->MYRI10GE_GSO_SIZE;
		max_segments = MYRI10GE_MAX_SEND_DESC_TSO;
	}

	if ((unlikely(avail < max_segments))) {
		/* we are out of transmit resources */
		mgp->stop_queue++;
		netif_stop_queue(dev);
		return 1;
	}

	/* Setup checksum offloading, if needed */
	cksum_offset = 0;
	pseudo_hdr_offset = 0;
	odd_flag = 0;
	flags = (MXGEFW_FLAGS_NO_TSO |
		 MXGEFW_FLAGS_FIRST);
	if (likely(skb->ip_summed == CHECKSUM_PARTIAL)) {
		cksum_offset = myri10ge_skb_transport_offset(skb);
		pseudo_hdr_offset = cksum_offset + skb->MYRI10GE_SKB_CSUM_OFFSET;
		/* If the headers are excessively large, then we must
		 * fall back to a software checksum */
		if (unlikely(cksum_offset > 255 ||
			     pseudo_hdr_offset > 127)) {
#ifdef LINUX_KERNEL_SPECIFIC
			if (skb_checksum_help(skb))
				goto drop;
#else  /* LINUX_KERNEL_SPECIFIC */
			myri10ge_csum_fixup(skb, cksum_offset, pseudo_hdr_offset);
#endif /* LINUX_KERNEL_SPECIFIC */
			cksum_offset = 0;
			pseudo_hdr_offset = 0;
		} else {
			odd_flag = MXGEFW_FLAGS_ALIGN_ODD;
			flags |= MXGEFW_FLAGS_CKSUM;
		}
	}

	cum_len = 0;

	if (mss) { /* TSO */
		/* this removes any CKSUM flag from before */
		flags = (MXGEFW_FLAGS_TSO_HDR |
			 MXGEFW_FLAGS_FIRST);

		/* negative cum_len signifies to the
		 * send loop that we are still in the
		 * header portion of the TSO packet.
		 * TSO header must be at most 134 bytes long */
		cum_len = -(myri10ge_skb_transport_offset(skb) + myri10ge_tcp_hdrlen(skb));

		/* for TSO, pseudo_hdr_offset holds mss.
		 * The firmware figures out where to put
		 * the checksum by parsing the header. */
		pseudo_hdr_offset = mss;
	} else
	/* Mark small packets, and pad out tiny packets */
	if (skb->len <= MXGEFW_SEND_SMALL_SIZE) {
		flags |= MXGEFW_FLAGS_SMALL;

		/* pad frames to at least ETH_ZLEN bytes */
		if (unlikely(skb->len < ETH_ZLEN)) {
			if (myri10ge_skb_padto(skb, ETH_ZLEN)) {
				/* The packet is gone, so we must
				   return 0 */
				mgp->stats.tx_dropped += 1;
				return 0;
			}
			/* adjust the len to account for the zero pad
			   so that the nic can know how long it is */
			skb->len = ETH_ZLEN;
		}
	}

	/* map the skb for DMA */
	len = skb->len - skb->data_len;
	idx = tx->req & tx->mask;
	tx->info[idx].skb = skb;
	bus = pci_map_single(mgp->pdev, skb->data, len, PCI_DMA_TODEVICE);
	pci_unmap_addr_set(&tx->info[idx], bus, bus);
	pci_unmap_len_set(&tx->info[idx], len, len);

	frag_cnt = skb_shinfo(skb)->nr_frags;
	frag_idx = 0;
	count = 0;
	rdma_count = 0;

	/* "rdma_count" is the number of RDMAs belonging to the
	 * current packet BEFORE the current send request. For
	 * non-TSO packets, this is equal to "count".
	 * For TSO packets, rdma_count needs to be reset
	 * to 0 after a segment cut.
	 *
	 * The rdma_count field of the send request is
	 * the number of RDMAs of the packet starting at
	 * that request. For TSO send requests with one ore more cuts
	 * in the middle, this is the number of RDMAs starting
	 * after the last cut in the request. All previous
	 * segments before the last cut implicitly have 1 RDMA.
	 *
	 * Since the number of RDMAs is not known beforehand,
	 * it must be filled-in retroactively - after each
	 * segmentation cut or at the end of the entire packet.
	 */

	while (1) {
		/* Break the SKB or Fragment up into pieces which
		   do not cross mgp->tx.boundary */
		low = MYRI10GE_LOWPART_TO_U32(bus);
		high_swapped = htonl(MYRI10GE_HIGHPART_TO_U32(bus));
		while (len) {
			u8 flags_next;
			int cum_len_next;

			if (unlikely(count == max_segments))
				goto abort_linearize;

			boundary = (low + tx->boundary) & ~(tx->boundary - 1);
			seglen = boundary - low;
			if (seglen > len)
				seglen = len;
			flags_next = flags & ~MXGEFW_FLAGS_FIRST;
			cum_len_next = cum_len + seglen;
			if (mss) { /* TSO */
				(req-rdma_count)->rdma_count = rdma_count + 1;

				if (likely(cum_len >= 0)) { /* payload */
					int next_is_first, chop;

					chop = (cum_len_next>mss);
					cum_len_next = cum_len_next % mss;
					next_is_first = (cum_len_next == 0);
					flags |= chop *
						MXGEFW_FLAGS_TSO_CHOP;
					flags_next |= next_is_first *
						MXGEFW_FLAGS_FIRST;
					rdma_count |= -(chop | next_is_first);
					rdma_count += chop & !next_is_first;
				} else if (likely(cum_len_next >= 0)) { /* header ends */
					int small;

					rdma_count = -1;
					cum_len_next = 0;
					seglen = -cum_len;
					small = (mss <= MXGEFW_SEND_SMALL_SIZE);
					flags_next = MXGEFW_FLAGS_TSO_PLD |
						MXGEFW_FLAGS_FIRST |
						(small * MXGEFW_FLAGS_SMALL);
				}
			}
			req->addr_high = high_swapped;
			req->addr_low = htonl(low);
			req->pseudo_hdr_offset = htons(pseudo_hdr_offset);
			req->pad = 0;	/* complete solid 16-byte block; does this matter? */
			req->rdma_count = 1;
			req->length = htons(seglen);
			req->cksum_offset = cksum_offset;
			req->flags = flags | ((cum_len & 1) * odd_flag);

			low += seglen;
			len -= seglen;
			cum_len = cum_len_next;
			flags = flags_next;
			req++;
			count++;
			rdma_count++;
			if (unlikely(cksum_offset > seglen))
				cksum_offset -= seglen;
			else
				cksum_offset = 0;
		}
		if (frag_idx == frag_cnt)
			break;

		/* map next fragment for DMA */
		idx = (count + tx->req) & tx->mask;
		frag = &skb_shinfo(skb)->frags[frag_idx];
		frag_idx++;
		len = frag->size;
		bus = pci_map_page(mgp->pdev, frag->page, frag->page_offset,
				   len, PCI_DMA_TODEVICE);
		pci_unmap_addr_set(&tx->info[idx], bus, bus);
		pci_unmap_len_set(&tx->info[idx], len, len);
	}

	(req-rdma_count)->rdma_count = rdma_count;
	if (mss)
		do {
			req--;
			req->flags |= MXGEFW_FLAGS_TSO_LAST;
		} while (!(req->flags & (MXGEFW_FLAGS_TSO_CHOP |
					 MXGEFW_FLAGS_FIRST)));
	idx = ((count - 1) + tx->req) & tx->mask;
	tx->info[idx].last = 1;
	if (tx->wc_fifo == NULL)
		myri10ge_submit_req(tx, tx->req_list, count);
	else
		myri10ge_submit_req_wc(tx, tx->req_list, count);
	tx->pkt_start++;
	if ((avail - count) < MXGEFW_MAX_SEND_DESC) {
		mgp->stop_queue++;
		netif_stop_queue(dev);
	}
	dev->trans_start = jiffies;
	return 0;


abort_linearize:
	/* Free any DMA resources we've alloced and clear out the skb
	 * slot so as to not trip up assertions, and to avoid a
	 * double-free if linearizing fails */

	last_idx = (idx + 1) & tx->mask;
	idx = tx->req & tx->mask;
	tx->info[idx].skb = NULL;
	do {
		len = pci_unmap_len(&tx->info[idx], len);
		if (len) {
			if (tx->info[idx].skb != NULL)
				pci_unmap_single(mgp->pdev,
						 pci_unmap_addr(&tx->info[idx], bus),
						 len, PCI_DMA_TODEVICE);
			else
				pci_unmap_page(mgp->pdev,
					       pci_unmap_addr(&tx->info[idx], bus),
					       len, PCI_DMA_TODEVICE);
			pci_unmap_len_set(&tx->info[idx], len, 0);
			tx->info[idx].skb = NULL;
		}
		idx = (idx + 1) & tx->mask;
	} while (idx != last_idx);
	if (myri10ge_skb_is_gso(skb)) {
		printk(KERN_ERR "myri10ge: %s: TSO but wanted to linearize?!?!?\n",
		       mgp->dev->name);
		goto drop;
	}

	if (myri10ge_skb_linearize(skb))
		goto drop;

	mgp->tx_linearized++;
	goto again;

drop:
	dev_kfree_skb_any(skb);
	mgp->stats.tx_dropped += 1;
	return 0;


}

static struct net_device_stats *
myri10ge_get_stats(struct net_device *dev)
{
	struct myri10ge_priv *mgp = netdev_priv(dev);
	return &mgp->stats;
}

static void
myri10ge_set_multicast_list(struct net_device *dev)
{
	struct myri10ge_cmd cmd;
	struct myri10ge_priv *mgp;
	struct dev_mc_list *mc_list;
	__be32 data[2] = {0, 0};
	int err;

	mgp = netdev_priv(dev);
	/* can be called from atomic contexts,
	 * pass 1 to force atomicity in myri10ge_send_cmd() */
	myri10ge_change_promisc(mgp, dev->flags & IFF_PROMISC, 1);

	/* This firmware is known to not support multicast */
	if (!mgp->fw_multicast_support)
		return;

	/* Disable multicast filtering */

	err = myri10ge_send_cmd(mgp, MXGEFW_ENABLE_ALLMULTI,
				&cmd, 1);
	if (err != 0) {
		printk(KERN_ERR "myri10ge: %s: Failed MXGEFW_ENABLE_ALLMULTI,"
		       " error status: %d\n", dev->name, err);
		goto abort;
	}

	if ((dev->flags & IFF_ALLMULTI) || mgp->adopted_rx_filter_bug) {
		/* request to disable multicast filtering, so quit here */
		return;
	}

	/* Flush the filters */

	err = myri10ge_send_cmd(mgp, MXGEFW_LEAVE_ALL_MULTICAST_GROUPS,
				&cmd, 1);
	if (err != 0) {
		printk(KERN_ERR
		       "myri10ge: %s: Failed MXGEFW_LEAVE_ALL_MULTICAST_GROUPS"
		       ", error status: %d\n", dev->name, err);
		goto abort;
	}

	/* Walk the multicast list, and add each address */
	for (mc_list = dev->mc_list; mc_list != NULL;
	     mc_list = mc_list->next) {
		memcpy(data, &mc_list->dmi_addr, 6);
		cmd.data0 = ntohl(data[0]);
		cmd.data1 = ntohl(data[1]);
		err = myri10ge_send_cmd(mgp, MXGEFW_JOIN_MULTICAST_GROUP,
					&cmd, 1);

		if (err != 0) {
			printk(KERN_ERR "myri10ge: %s: Failed "
			       "MXGEFW_JOIN_MULTICAST_GROUP, error status:"
			       "%d\t", dev->name, err);
			printk(KERN_ERR "MAC %02x:%02x:%02x:%02x:%02x:%02x\n",
			       ((unsigned char *)&mc_list->dmi_addr)[0],
			       ((unsigned char *)&mc_list->dmi_addr)[1],
			       ((unsigned char *)&mc_list->dmi_addr)[2],
			       ((unsigned char *)&mc_list->dmi_addr)[3],
			       ((unsigned char *)&mc_list->dmi_addr)[4],
			       ((unsigned char *)&mc_list->dmi_addr)[5]
			       );
			goto abort;
		}
	}
	/* Enable multicast filtering */
	err = myri10ge_send_cmd(mgp, MXGEFW_DISABLE_ALLMULTI,
				&cmd, 1);
	if (err != 0) {
		printk(KERN_ERR "myri10ge: %s: Failed MXGEFW_DISABLE_ALLMULTI,"
		       "error status: %d\n", dev->name, err);
		goto abort;
	}

	return;

  abort:
	return;
}


static int
myri10ge_set_mac_address (struct net_device *dev, void *addr)
{
	struct sockaddr *sa = addr;
	struct myri10ge_priv *mgp = netdev_priv(dev);
	int status;

	if (!is_valid_ether_addr(sa->sa_data))
		return -EADDRNOTAVAIL;

	status = myri10ge_update_mac_address(mgp, sa->sa_data);
	if (status != 0) {
		printk(KERN_ERR "myri10ge: %s: changing mac address failed with %d\n",
		       dev->name, status);
		return status;
	}

	/* change the dev structure */
	memcpy(dev->dev_addr, sa->sa_data, 6);
	return 0;
}

static int
myri10ge_change_mtu(struct net_device *dev, int new_mtu)
{
	struct myri10ge_priv *mgp = netdev_priv(dev);
	int error = 0;

	if ((new_mtu < 68) || (ETH_HLEN + new_mtu > MYRI10GE_MAX_ETHER_MTU)) {
		printk(KERN_ERR "myri10ge: %s: new mtu (%d) is not valid\n",
		       dev->name, new_mtu);
		return -EINVAL;
	}
	printk(KERN_INFO "%s: changing mtu from %d to %d\n",
	       dev->name, dev->mtu, new_mtu);
	if (mgp->running) {
		/* if we change the mtu on an active device, we must
		 * reset the device so the firmware sees the change */
		myri10ge_close(dev);
		dev->mtu = new_mtu;
		myri10ge_open(dev);
	} else
		dev->mtu = new_mtu;

	return error;
}

/*
 * Enable ECRC to align PCI-E Completion packets on an 8-byte boundary.
 * Only do it if the bridge is a root port since we don't want to disturb
 * any other device, except if forced with myri10ge_ecrc_enable > 1.
 */

static void
myri10ge_enable_ecrc(struct myri10ge_priv *mgp)
{
	struct pci_dev *bridge = mgp->pdev->bus->self;
	struct device * dev = &mgp->pdev->dev;
	unsigned cap;
	unsigned err_cap;
	u16 val;
	u8 ext_type;
	int ret;

	if (!myri10ge_ecrc_enable || !bridge)
		return;

	/* check that the bridge is a root port */
	cap = pci_find_capability(bridge, PCI_CAP_ID_EXP);
	pci_read_config_word(bridge, cap + PCI_CAP_FLAGS, &val);
	ext_type = (val & PCI_EXP_FLAGS_TYPE) >> 4;
	if (ext_type != PCI_EXP_TYPE_ROOT_PORT) {
		if (myri10ge_ecrc_enable > 1) {
			struct pci_dev * old_bridge = bridge;

			/* Walk the hierarchy up to the root port
			 * where ECRC has to be enabled */
			do {
				bridge = bridge->bus->self;
				if (!bridge) {
					dev_err(dev,
						"Failed to find root port"
						" to force ECRC\n");
					return;
				}
				cap = pci_find_capability(bridge, PCI_CAP_ID_EXP);
				pci_read_config_word(bridge, cap + PCI_CAP_FLAGS, &val);
				ext_type = (val & PCI_EXP_FLAGS_TYPE) >> 4;
			} while (ext_type != PCI_EXP_TYPE_ROOT_PORT);

			dev_info(dev,
				 "Forcing ECRC on non-root port %s"
				 " (enabling on root port %s)\n",
				 pci_name(old_bridge), pci_name(bridge));
		} else {
			dev_err(dev,
				"Not enabling ECRC on non-root port %s\n",
				pci_name(bridge));
			return;
		}
	}

	cap = pci_find_ext_capability(bridge, PCI_EXT_CAP_ID_ERR);
#ifndef LINUX_KERNEL_SPECIFIC
	/* nvidia ext cap is not always linked in ext cap chain,
	 * fixed with a quirk in 2.6.18 */
	if (!cap
	    && bridge->vendor == PCI_VENDOR_ID_NVIDIA
	    && (bridge->device == PCI_DEVICE_ID_NVIDIA_NFORCE_CK804_PCIE
		|| (bridge->device >= PCI_DEVICE_ID_NVIDIA_NFORCE_MCP55_PCIE_374 
		    && bridge->device <= PCI_DEVICE_ID_NVIDIA_NFORCE_MCP55_PCIE_378)))
		cap = 0x160;

#endif
	if (!cap)
		return;

#ifdef LINUX_KERNEL_SPECIFIC
	ret = pci_read_config_dword(bridge, cap + PCI_ERR_CAP, &err_cap);
#else /* LINUX_KERNEL_SPECIFIC */
	ret = myri10ge_read_ext_config_dword(bridge, cap + PCI_ERR_CAP, &err_cap);
#endif /* LINUX_KERNEL_SPECIFIC */
	if (ret) {
		dev_err(dev, "failed reading ext-conf-space of %s\n",
			pci_name(bridge));
		dev_err(dev, "\t pci=nommconf in use? "
			"or buggy/incomplete/absent ACPI MCFG attr?\n");
		return;
	}
	if (!(err_cap & PCI_ERR_CAP_ECRC_GENC))
		return;

	err_cap |= PCI_ERR_CAP_ECRC_GENE;
#ifdef LINUX_KERNEL_SPECIFIC
	pci_write_config_dword(bridge, cap + PCI_ERR_CAP, err_cap);
#else /* LINUX_KERNEL_SPECIFIC */
	myri10ge_write_ext_config_dword(bridge, cap + PCI_ERR_CAP, err_cap);
#endif /* LINUX_KERNEL_SPECIFIC */
	dev_info(dev,
		 "Enabled ECRC on upstream bridge %s\n",
		 pci_name(bridge));
}

/*
 * The Lanai Z8E PCI-E interface achieves higher Read-DMA throughput
 * when the PCI-E Completion packets are aligned on an 8-byte
 * boundary.  Some PCI-E chip sets always align Completion packets; on
 * the ones that do not, the alignment can be enforced by enabling
 * ECRC generation (if supported).
 *
 * When PCI-E Completion packets are not aligned, it is actually more
 * efficient to limit Read-DMA transactions to 2KB, rather than 4KB.
 *
 * If the driver can neither enable ECRC nor verify that it has
 * already been enabled, then it must use a firmware image which works
 * around unaligned completion packets (myri10ge_ethp_z8e.dat), and it
 * should also ensure that it never gives the device a Read-DMA which is
 * larger than 2KB by setting the tx.boundary to 2KB.  If ECRC is
 * enabled, then the driver should use the aligned (myri10ge_eth_z8e.dat)
 * firmware image, and set tx.boundary to 4KB.
 */

static void
myri10ge_firmware_probe(struct myri10ge_priv *mgp)
{
	struct pci_dev *pdev = mgp->pdev;
	struct device *dev = &pdev->dev;
	int cap, status;
	u16 val;

	mgp->tx.boundary = 4096;
	/*
	 * Verify the max read request size was set to 4KB
	 * before trying the test with 4KB.
	 */
	cap = pci_find_capability(pdev, PCI_CAP_ID_EXP);
	if (cap < 64) {
		dev_err(dev, "Bad PCI_CAP_ID_EXP location %d\n", cap);
		goto abort;
	}
	status = pci_read_config_word(pdev, cap + PCI_EXP_DEVCTL, &val);
	if (status != 0) {
		dev_err(dev, "Couldn't read max read req size: %d\n", status);
		goto abort;
	}
	if ((val & (5 << 12)) != (5 << 12)) {
		dev_warn(dev, "Max Read Request size != 4096 (0x%x)\n", val);
		mgp->tx.boundary = 2048;
	}
	/* 
	 * load the optimized firmware (which assumes aligned PCIe
	 * completions) in order to see if it works on this host.
	 */
	mgp->fw_name = myri10ge_fw_aligned;
	status = myri10ge_load_firmware(mgp);
	if (status != 0) {
		goto abort;
	}

	/* 
	 * Enable ECRC if possible
	 */
	myri10ge_enable_ecrc(mgp);

	/* 
	 * Run a DMA test which watches for unaligned completions and
	 * aborts on the first one seen.
	 */

	status = myri10ge_dma_test(mgp, MXGEFW_CMD_UNALIGNED_TEST);
	if (status == 0)
		return; /* keep the aligned firmware */

	if (status != -E2BIG)
		dev_warn(dev, "DMA test failed: %d\n", status);
	if (status == -ENOSYS)
		dev_warn(dev, "Falling back to ethp! "
			 "Please install up to date fw\n");
abort:
	/* fall back to using the unaligned firmware */
	mgp->tx.boundary = 2048;
	mgp->fw_name = myri10ge_fw_unaligned;
	
	
}

static void
myri10ge_select_firmware(struct myri10ge_priv *mgp)
{
	if (myri10ge_force_firmware == 0) {
		int link_width, exp_cap;
		u16 lnk;

		exp_cap = pci_find_capability(mgp->pdev, PCI_CAP_ID_EXP);
		pci_read_config_word(mgp->pdev, exp_cap + PCI_EXP_LNKSTA, &lnk);
		link_width = (lnk >> 4) & 0x3f;

		/* Check to see if Link is less than 8 or if the
		 * upstream bridge is known to provide aligned
		 * completions */
		if (link_width < 8) {
			dev_info(&mgp->pdev->dev, "PCIE x%d Link\n", link_width);
			mgp->tx.boundary = 4096;
			mgp->fw_name = myri10ge_fw_aligned;
		} else { 
			myri10ge_firmware_probe(mgp);
		}
	} else {
		if (myri10ge_force_firmware == 1) {
			dev_info(&mgp->pdev->dev,
				 "Assuming aligned completions (forced)\n");
			mgp->tx.boundary = 4096;
			mgp->fw_name = myri10ge_fw_aligned;
		} else {
			dev_info(&mgp->pdev->dev,
				 "Assuming unaligned completions (forced)\n");
			mgp->tx.boundary = 2048;
			mgp->fw_name = myri10ge_fw_unaligned;
		}
	}
	if (myri10ge_fw_name != NULL) {
		dev_info(&mgp->pdev->dev, "overriding firmware to %s\n",
			 myri10ge_fw_name);
		mgp->fw_name = myri10ge_fw_name;
	}
}


#ifndef LINUX_KERNEL_SPECIFIC
static void
myri10ge_save_state(struct pci_dev *pdev)
{
	struct myri10ge_priv *mgp;
	int cap;

	mgp = pci_get_drvdata(pdev);
	BUG_ON(mgp == NULL);

	pci_save_state(pdev);
	/* now save PCIe and MSI state that Linux will not
	   save for us */
	cap = pci_find_capability(pdev, PCI_CAP_ID_EXP);
	pci_read_config_dword(pdev, cap + PCI_EXP_DEVCTL, &mgp->devctl);
	cap = pci_find_capability(pdev, PCI_CAP_ID_MSI);
	pci_read_config_word(pdev, cap + PCI_MSI_FLAGS, &mgp->msi_flags);
	pci_read_config_dword(pdev, cap + PCI_MSI_ADDRESS_LO,
			      &mgp->msi_addr_low);
	pci_read_config_dword(pdev, cap + PCI_MSI_ADDRESS_HI,
			      &mgp->msi_addr_high);
	pci_read_config_word(pdev, cap + PCI_MSI_DATA_32,
			     &mgp->msi_data_32);
	pci_read_config_word(pdev, cap + PCI_MSI_DATA_64,
			     &mgp->msi_data_64);
}

static int
myri10ge_restore_state(struct pci_dev *pdev)
{
	struct myri10ge_priv *mgp;
	int cap;

	mgp = pci_get_drvdata(pdev);
	BUG_ON(mgp == NULL);

	/* restore PCIe and MSI state that linux will not */
	cap = pci_find_capability(pdev, PCI_CAP_ID_EXP);
	pci_write_config_dword(pdev, cap + PCI_CAP_ID_EXP, mgp->devctl);
	cap = pci_find_capability(pdev, PCI_CAP_ID_MSI);
	pci_write_config_word(pdev, cap + PCI_MSI_FLAGS, mgp->msi_flags);
	pci_write_config_dword(pdev, cap + PCI_MSI_ADDRESS_LO,
			       mgp->msi_addr_low);
	pci_write_config_dword(pdev, cap + PCI_MSI_ADDRESS_HI,
			       mgp->msi_addr_high);
	pci_write_config_word(pdev, cap + PCI_MSI_DATA_32,
			      mgp->msi_data_32);
	pci_write_config_word(pdev, cap + PCI_MSI_DATA_64,
			      mgp->msi_data_64);

	return pci_restore_state(pdev);
}
#endif /* LINUX_KERNEL_SPECIFIC */

#ifdef CONFIG_PM

static int
myri10ge_suspend(struct pci_dev *pdev, pm_message_t state)
{
	struct myri10ge_priv *mgp;
	struct net_device *netdev;

	mgp = pci_get_drvdata(pdev);
	if (mgp == NULL)
		return -EINVAL;
	netdev = mgp->dev;

	netif_device_detach(netdev);
	if (netif_running(netdev)) {
		printk(KERN_INFO "myri10ge: closing %s\n", netdev->name);
		rtnl_lock();
		myri10ge_close(netdev);
		rtnl_unlock();
	}
	myri10ge_dummy_rdma(mgp, 0);
	myri10ge_save_state(pdev);
	pci_disable_device(pdev);

	return pci_set_power_state(pdev, pci_choose_state(pdev, state));
}

static int
myri10ge_resume(struct pci_dev *pdev)
{
	struct myri10ge_priv *mgp;
	struct net_device *netdev;
	int status;
	u16 vendor;

	mgp = pci_get_drvdata(pdev);
	if (mgp == NULL)
		return -EINVAL;
	netdev = mgp->dev;
	pci_set_power_state(pdev, 0);  /* zeros conf space as a side effect */
	myri10ge_msleep(5);	/* give card time to respond */
	pci_read_config_word(mgp->pdev, PCI_VENDOR_ID, &vendor);
	if (vendor == 0xffff) {
		printk(KERN_ERR "myri10ge: %s: device disappeared!\n",
		       mgp->dev->name);
		return -EIO;
	}

	status = myri10ge_restore_state(pdev);
	if (status)
		return status;

	status = pci_enable_device(pdev);
	if (status) {
		dev_err(&pdev->dev, "failed to enable device\n");
		return status;
	}

	pci_set_master(pdev);

	myri10ge_reset(mgp);
	myri10ge_dummy_rdma(mgp, 1);

	/* Save configuration space to be restored if the
	   nic resets due to a parity error */
	myri10ge_save_state(pdev);

	if (netif_running(netdev)) {
		rtnl_lock();
		status = myri10ge_open(netdev);
		rtnl_unlock();
		if (status != 0)
			goto abort_with_enabled;
		
	}
	netif_device_attach(netdev);

	return 0;

abort_with_enabled:
	pci_disable_device(pdev);
	return -EIO;

}

#endif /* CONFIG_PM */

static u32
myri10ge_read_reboot(struct myri10ge_priv *mgp)
{
	struct pci_dev *pdev = mgp->pdev;
	int vs = mgp->vendor_specific_offset;
	u32 reboot;

	/*enter read32 mode */
	pci_write_config_byte(pdev, vs + 0x10, 0x3);

	/*read REBOOT_STATUS (0xfffffff0) */
	pci_write_config_dword(pdev, vs + 0x18, 0xfffffff0);
	pci_read_config_dword(pdev, vs + 0x14, &reboot);
	return reboot;
}

/*
 * This watchdog is used to check whether the board has suffered
 * from a parity error and needs to be recovered.
 */
static void
myri10ge_watchdog(MYRI10GE_WATCHDOG_ARG_TYPE work)
{
	struct myri10ge_priv *mgp = MYRI10GE_WATCHDOG_ARG_CONTAINER_OF_MGP(work, struct myri10ge_priv, watchdog_work);
	u32 reboot;
	int status;
	u16 cmd, vendor;

	mgp->watchdog_resets++;
	pci_read_config_word(mgp->pdev, PCI_COMMAND, &cmd);
	if ((cmd & PCI_COMMAND_MASTER) == 0) {
		/* Bus master DMA disabled?  Check to see
		 * if the card rebooted due to a parity error
		 * For now, just report it */
		reboot = myri10ge_read_reboot(mgp);
		printk(KERN_ERR "myri10ge: %s: NIC rebooted (0x%x),%s resetting\n",
		       mgp->dev->name, reboot, 
		       myri10ge_reset_recover ? " " : " not");
		if (myri10ge_reset_recover == 0)
			return;

		myri10ge_reset_recover--;
		
		/*
		 * A rebooted nic will come back with config space as
		 * it was after power was applied to PCIe bus.
		 * Attempt to restore config space which was saved
		 * when the driver was loaded, or the last time the
		 * nic was resumed from power saving mode.
		 */
		myri10ge_restore_state(mgp->pdev);

		/* save state again for accounting reasons */
		myri10ge_save_state(mgp->pdev);

	} else {
		/* if we get back -1's from our slot, perhaps somebody
		   powered off our card.  Don't try to reset it in
		   this case */
		if (cmd == 0xffff) {
			pci_read_config_word(mgp->pdev, PCI_VENDOR_ID, &vendor);
			if (vendor == 0xffff) {
				printk(KERN_ERR "myri10ge: %s: device disappeared!\n",
				       mgp->dev->name);
				return;
			}
		}
		/* Perhaps it is a software error.  Try to reset */

		printk(KERN_ERR "myri10ge: %s: device timeout, resetting\n",
		       mgp->dev->name);
		printk(KERN_INFO "myri10ge: %s: %d %d %d %d %d\n", mgp->dev->name,
		       mgp->tx.req, mgp->tx.done, mgp->tx.pkt_start,
		       mgp->tx.pkt_done,
		       (int)ntohl(mgp->fw_stats->send_done_count));
		myri10ge_msleep(2000);
		printk(KERN_INFO "myri10ge: %s: %d %d %d %d %d\n", mgp->dev->name,
		       mgp->tx.req, mgp->tx.done, mgp->tx.pkt_start,
		       mgp->tx.pkt_done,
		       (int)ntohl(mgp->fw_stats->send_done_count));
	}
	rtnl_lock();
	myri10ge_close(mgp->dev);
	status = myri10ge_load_firmware(mgp);
	if (status != 0)
		printk(KERN_ERR "myri10ge: %s: failed to load firmware\n",
		       mgp->dev->name);
	else
		myri10ge_open(mgp->dev);
	rtnl_unlock();
}

/*
 * We use our own timer routine rather than relying upon
 * netdev->tx_timeout because we have a very large hardware transmit
 * queue.  Due to the large queue, the netdev->tx_timeout function
 * cannot detect a NIC with a parity error in a timely fashion if the
 * NIC is lightly loaded.
 */
static void
myri10ge_watchdog_timer(unsigned long arg)
{
	struct myri10ge_priv *mgp;

	mgp = (struct myri10ge_priv *) arg;

#if !MYRI10GE_RX_SKBS
	if (mgp->rx_small.watchdog_needed) {
		myri10ge_alloc_rx_pages(mgp, &mgp->rx_small,
					mgp->small_bytes + MXGEFW_PAD, 1);
		if (mgp->rx_small.fill_cnt -  mgp->rx_small.cnt >= myri10ge_fill_thresh)
			mgp->rx_small.watchdog_needed = 0;
	}
	if (mgp->rx_big.watchdog_needed) {
		myri10ge_alloc_rx_pages(mgp, &mgp->rx_big,
					mgp->big_bytes, 1);
		if (mgp->rx_big.fill_cnt -  mgp->rx_big.cnt >= myri10ge_fill_thresh)
			mgp->rx_big.watchdog_needed = 0;
	}
#endif /* MYRI10GE_RX_SKBS */

	if (mgp->tx.req != mgp->tx.done &&
	    mgp->tx.done == mgp->watchdog_tx_done &&
	    mgp->watchdog_tx_req != mgp->watchdog_tx_done)
		/* nic seems like it might be stuck.. */
		schedule_work(&mgp->watchdog_work);
	else
		/* rearm timer */
		mod_timer(&mgp->watchdog_timer,
			  jiffies + myri10ge_watchdog_timeout * HZ);

	mgp->watchdog_tx_done = mgp->tx.done;
	mgp->watchdog_tx_req = mgp->tx.req;
}

static int
myri10ge_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	struct net_device *netdev;
	struct myri10ge_priv *mgp;
	struct device *dev = &pdev->dev;
	size_t bytes;
	int i;
	int status = -ENXIO;
	int cap;
	int dac_enabled;
	u16 val;

#ifndef LINUX_KERNEL_SPECIFIC
	if (myri10ge_bus != -1 && pdev->bus->number != myri10ge_bus)
		return -ENODEV;
#endif

	netdev = alloc_etherdev(sizeof(*mgp));
	if (netdev == NULL) {
		dev_err(dev, "Could not allocate ethernet device\n");
		return -ENOMEM;
	}

	SET_NETDEV_DEV(netdev, &pdev->dev);

	mgp = netdev_priv(netdev);
	memset(mgp, 0, sizeof (*mgp));
	mgp->dev = netdev;
	mgp->pdev = pdev;
	mgp->csum_flag = MXGEFW_FLAGS_CKSUM;
	mgp->pause = myri10ge_flow_control;
	mgp->intr_coal_delay = myri10ge_intr_coal_delay;
	mgp->msg_enable = netif_msg_init(myri10ge_debug, MYRI10GE_MSG_DEFAULT);
	init_waitqueue_head(&mgp->down_wq);

	if (pci_enable_device(pdev)) {
		dev_err(&pdev->dev, "pci_enable_device call failed\n");
		status = -ENODEV;
		goto abort_with_netdev;
	}

	/* Find the vendor-specific cap so we can check
	   the reboot register later on */
	mgp->vendor_specific_offset
		= pci_find_capability(pdev, PCI_CAP_ID_VNDR);

	/* Set our max read request to 4KB */
	cap = pci_find_capability(pdev, PCI_CAP_ID_EXP);
	if (cap < 64) {
		dev_err(&pdev->dev,"Bad PCI_CAP_ID_EXP location %d\n", cap);
		goto abort_with_netdev;
	}
	status = pci_read_config_word(pdev, cap + PCI_EXP_DEVCTL, &val);
	if (status != 0) {
		dev_err(&pdev->dev, "Error %d reading PCI_EXP_DEVCTL\n", status);
		goto abort_with_netdev;
	}
	val = (val & ~PCI_EXP_DEVCTL_READRQ) | (5 << 12);
	status = pci_write_config_word(pdev, cap + PCI_EXP_DEVCTL, val);
	if (status != 0) {
		dev_err(&pdev->dev, "Error %d writing PCI_EXP_DEVCTL\n", status);
		goto abort_with_netdev;
	}

	pci_set_master(pdev);
	dac_enabled = 1;
	status = pci_set_dma_mask(pdev, DMA_64BIT_MASK);
	if (status != 0) {
		dac_enabled = 0;
		dev_err(&pdev->dev, "64-bit pci address mask was refused, trying 32-bit");
		status = pci_set_dma_mask(pdev, DMA_32BIT_MASK);
	}
	if (status != 0) {
		dev_err(&pdev->dev, "Error %d setting DMA mask\n", status);
		goto abort_with_netdev;
	}
	mgp->cmd = dma_alloc_coherent(&pdev->dev, sizeof (*mgp->cmd),
				      &mgp->cmd_bus, GFP_KERNEL);
	if (mgp->cmd == NULL)
		goto abort_with_netdev;

	mgp->fw_stats = dma_alloc_coherent(&pdev->dev, sizeof (*mgp->fw_stats),
					   &mgp->fw_stats_bus, GFP_KERNEL);
	if (mgp->fw_stats == NULL)
		goto abort_with_cmd;

	mgp->board_span = pci_resource_len(pdev, 0);
	mgp->iomem_base = pci_resource_start(pdev, 0);
	mgp->mtrr = -1;
	mgp->wc_enabled = 0;
#if MYRI10GE_HAVE_PAT
	if (mgp->mtrr < 0) {
		on_each_cpu(myri10ge_enable_pat, 0, 0, 1);
	}
	if (!myri10ge_pat_failed)
		mgp->wc_enabled = 2;
#endif
#ifdef CONFIG_MTRR
#ifndef LINUX_KERNEL_SPECIFIC
	if (!mgp->wc_enabled)
#endif /* LINUX_KERNEL_SPECIFIC */
	mgp->mtrr = mtrr_add(mgp->iomem_base, mgp->board_span,
			     MTRR_TYPE_WRCOMB, 1);
	if (mgp->mtrr >= 0)
		mgp->wc_enabled = 1;
#endif
	/* Hack.  need to get rid of these magic numbers */
	mgp->sram_size = 2*1024*1024 - (2*(48*1024)+(32*1024)) - 0x100;
	if (mgp->sram_size > mgp->board_span) {
		dev_err(&pdev->dev, "board span %ld bytes too small\n",
		       mgp->board_span);
		goto abort_with_wc;
	}
#if MYRI10GE_HAVE_PAT
	if (mgp->wc_enabled == 2)
		mgp->sram = __ioremap(mgp->iomem_base, mgp->board_span,
				      _PAGE_PCD | _PAGE_PSE);
	else
#endif
	mgp->sram = ioremap(mgp->iomem_base, mgp->board_span);
	if (mgp->sram == NULL) {
		dev_err(&pdev->dev, "ioremap failed for %ld bytes at 0x%lx\n",
			mgp->board_span, mgp->iomem_base);
		status = -ENXIO;
		goto abort_with_wc;
	}
	memcpy_fromio(mgp->eeprom_strings,
		      mgp->sram + mgp->sram_size - MYRI10GE_EEPROM_STRINGS_SIZE,
		      MYRI10GE_EEPROM_STRINGS_SIZE);
	memset(mgp->eeprom_strings + MYRI10GE_EEPROM_STRINGS_SIZE - 2, 0, 2);
	status = myri10ge_read_mac_addr(mgp);
	if (status)
		goto abort_with_ioremap;

	for (i = 0; i < ETH_ALEN; i++)
		netdev->dev_addr[i] = mgp->mac_addr[i];

	/* allocate rx done ring */
	bytes = myri10ge_max_intr_slots * sizeof (*mgp->rx_done.entry);
	mgp->rx_done.entry = dma_alloc_coherent(&pdev->dev, bytes,
						&mgp->rx_done.bus, GFP_KERNEL);
	if (mgp->rx_done.entry == NULL)
		goto abort_with_ioremap;
	memset(mgp->rx_done.entry, 0, bytes);

	myri10ge_select_firmware(mgp);

	status = myri10ge_load_firmware(mgp);
	if (status != 0) {
		dev_err(&pdev->dev, "failed to load firmware\n");
		goto abort_with_rx_done;
	}

	status = myri10ge_reset(mgp);
	if (status != 0) {
		dev_err(&pdev->dev, "failed reset\n");
		goto abort_with_firmware;
	}

	pci_set_drvdata(pdev, mgp);
	if ((myri10ge_initial_mtu + ETH_HLEN) > MYRI10GE_MAX_ETHER_MTU)
		myri10ge_initial_mtu = MYRI10GE_MAX_ETHER_MTU - ETH_HLEN;
	if ((myri10ge_initial_mtu + ETH_HLEN) < 68)
		myri10ge_initial_mtu = 68;
	netdev->mtu = myri10ge_initial_mtu;
	netdev->open = myri10ge_open;
	netdev->stop = myri10ge_close;
	netdev->hard_start_xmit = myri10ge_xmit;
	netdev->get_stats = myri10ge_get_stats;
	netdev->base_addr = mgp->iomem_base;
	netdev->change_mtu = myri10ge_change_mtu;
	netdev->set_multicast_list = myri10ge_set_multicast_list;
	netdev->set_mac_address = myri10ge_set_mac_address;
	netdev->features = NETIF_F_SG | NETIF_F_HW_CSUM | NETIF_F_TSO;
	if (dac_enabled)
		netdev->features |= NETIF_F_HIGHDMA;
#ifdef MYRI10GE_NAPI
	netdev->poll = myri10ge_poll;
	netdev->weight = myri10ge_napi_weight;
#endif

	/* make sure we can get an irq, and that MSI can be
	   setup (if available).  Also ensure netdev->irq
	   is set to correct value if MSI is enabled */
	status = myri10ge_request_irq(mgp);
	if (status != 0)
		goto abort_with_firmware;
	netdev->irq = pdev->irq;
	myri10ge_free_irq(mgp);

	/* Save configuration space to be restored if the
	 * nic resets due to a parity error */
	myri10ge_save_state(pdev);

	/* Setup the watchdog timer */
#ifdef LINUX_KERNEL_SPECIFIC
	setup_timer(&mgp->watchdog_timer, myri10ge_watchdog_timer, (unsigned long) mgp);
#else /* LINUX_KERNEL_SPECIFIC */
	/* setup_timer appeared in 2.6.15, no real need to add a HAL for 3 lines */
	init_timer(&mgp->watchdog_timer);
	mgp->watchdog_timer.data = (unsigned long)mgp;
	mgp->watchdog_timer.function = myri10ge_watchdog_timer;
#endif /* LINUX_KERNEL_SPECIFIC */

	SET_ETHTOOL_OPS(netdev, &myri10ge_ethtool_ops);
	MYRI10GE_INIT_WORK(&mgp->watchdog_work, myri10ge_watchdog, mgp);
	status = register_netdev(netdev);
	if (status != 0) {
		dev_err(&pdev->dev, "register_netdev failed: %d\n", status);
		goto abort_with_state;
	}
	dev_info(dev, "%s IRQ %d, tx bndry %d, fw %s, WC %s\n",
		 (mgp->msi_enabled ? "MSI" : "xPIC"),
		 netdev->irq, mgp->tx.boundary, mgp->fw_name,
		 (mgp->wc_enabled ? "Enabled" : "Disabled"));

	return 0;

abort_with_state:
	myri10ge_restore_state(pdev);

abort_with_firmware:
	myri10ge_dummy_rdma(mgp, 0);

abort_with_rx_done:
	bytes = myri10ge_max_intr_slots * sizeof (*mgp->rx_done.entry);
	dma_free_coherent(&pdev->dev, bytes,
			  mgp->rx_done.entry, mgp->rx_done.bus);

abort_with_ioremap:
	myri10ge_cleanup_linear_map(mgp);
	iounmap(mgp->sram);

abort_with_wc:
#ifdef CONFIG_MTRR
	if (mgp->mtrr >= 0)
		mtrr_del(mgp->mtrr, mgp->iomem_base, mgp->board_span);
#endif
	dma_free_coherent(&pdev->dev, sizeof (*mgp->fw_stats),
			  mgp->fw_stats, mgp->fw_stats_bus);

abort_with_cmd:
	dma_free_coherent(&pdev->dev, sizeof (*mgp->cmd),
			  mgp->cmd, mgp->cmd_bus);

abort_with_netdev:

	free_netdev(netdev);
	return status;
}

/*
 * myri10ge_remove
 *
 * Does what is necessary to shutdown one Myrinet device. Called
 *   once for each Myrinet card by the kernel when a module is
 *   unloaded.
 */
static void
myri10ge_remove(struct pci_dev *pdev)
{
	struct myri10ge_priv *mgp;
	struct net_device *netdev;
	size_t bytes;

	mgp = pci_get_drvdata(pdev);
	if (mgp == NULL)
		return;

	flush_scheduled_work();
	netdev = mgp->dev;
	unregister_netdev(netdev);

	myri10ge_dummy_rdma(mgp, 0);

	/* avoid a memory leak */
	myri10ge_restore_state(pdev);

	bytes = myri10ge_max_intr_slots * sizeof (*mgp->rx_done.entry);
	dma_free_coherent(&pdev->dev, bytes,
			  mgp->rx_done.entry, mgp->rx_done.bus);

	myri10ge_cleanup_linear_map(mgp);
	iounmap(mgp->sram);

#ifdef CONFIG_MTRR
	if (mgp->mtrr >= 0)
		mtrr_del(mgp->mtrr, mgp->iomem_base, mgp->board_span);
#endif
	dma_free_coherent(&pdev->dev, sizeof (*mgp->fw_stats),
			    mgp->fw_stats, mgp->fw_stats_bus);

	dma_free_coherent(&pdev->dev, sizeof (*mgp->cmd),
			  mgp->cmd, mgp->cmd_bus);

	free_netdev(netdev);
	pci_set_drvdata(pdev, NULL);
}


#define PCI_DEVICE_ID_MYRICOM_MYRI10GE_Z8E 	0x0008

static struct pci_device_id myri10ge_pci_tbl[] = {
	{ PCI_DEVICE(PCI_VENDOR_ID_MYRICOM, PCI_DEVICE_ID_MYRICOM_MYRI10GE_Z8E) },
	{ 0 },
};

static struct pci_driver myri10ge_driver = {
	.name = "myri10ge",
	.probe = myri10ge_probe,
	.remove = myri10ge_remove,
	.id_table = myri10ge_pci_tbl,
#ifdef CONFIG_PM
	.suspend = myri10ge_suspend,
	.resume = myri10ge_resume,
#endif
};

static __init int
myri10ge_init_module(void)
{
	printk(KERN_INFO "%s: Version %s\n", myri10ge_driver.name,
	       MYRI10GE_VERSION_STR);
#ifdef LINUX_KERNEL_SPECIFIC
	return pci_register_driver(&myri10ge_driver);	
#else /* LINUX_KERNEL_SPECIFIC */
#if !MYRI10GE_LRO
	if (myri10ge_lro != 0) {
		printk(KERN_ERR "%s: non-zero myri10ge_lro ignored, LRO was disabled at compile time\n",
			myri10ge_driver.name);
		myri10ge_lro = 0;
	}
#endif
	{
		int rc;
		rc = pci_register_driver(&myri10ge_driver);
		return rc < 0 ? rc : 0;
	}
#endif /* LINUX_KERNEL_SPECIFIC */
}
module_init(myri10ge_init_module);

static __exit void
myri10ge_cleanup_module(void)
{
	pci_unregister_driver(&myri10ge_driver);
}
module_exit(myri10ge_cleanup_module);

#ifndef LINUX_KERNEL_SPECIFIC
/*
  This file uses Myri10GE driver indentation.

  Local Variables:
  c-file-style:"linux"
  tab-width:8
  End:
 */
#endif /* LINUX_KERNEL_SPECIFIC */
